Attribute VB_Name = "LEF_Report_2014_11_20"
Option Explicit
Option Compare Text ' Non-case sensitive

'   * Copyright  2009-2014 by John P. Desautels (John@SmartBizNiz.com)
'   * Permission is granted to any individual/institution/corporate entity to use, copy, redistribute
'   * or modify this software for any purpose whatsoever, subject to one condition: namely, that this
'   * copyright block, in its entirety, not be removed from the source code.
'   *
'   * Source code, samples and documentation can be found at http://www.SmartBizNiz.com/downloads.html
'   *
'   * Special thanks to:
'   *   Ted Rau who introduced me to Launch Excel and inspired me to develop and share this framework
'   *   Guy Brown who helped me as I developed intricate YSL reports, which are the genesis of this framework
'   *   Deborah Brown, my first Yardi report writing instructor, who explained the history behind Launch Excel
'   *   Geoffrey Gelb who provided interesting projects which have allowed me to enhance the framework in support
'   *       of their requirements
'   *   All the Yardi VBA developers who have come before me, paving the way for this framework

Const conLEFVersion As String = "2014_11_20" ' This is the current version of the Launch Excel Framework
Public boolNewWorkbook As Boolean            ' Determines if report sheets are copied to a new workbook
Public boolSaveAs As Boolean                 ' Determines if the SaveAs dialog is displayed
Public boolProtectSheets As Boolean          ' Determines if the worksheets are to be protected after creation
Public myPassword As String                  ' If boolProtectSheets = True then this is the password to use
Public boolDisplayCSV As Boolean             ' Determines if the CSV worksheet is displayed in the final workbook
Public boolDeleteOldCSV As Boolean           ' Determines if we delete CSV files more than X hours old from \Local Settings\Application Data\YSI.NetClient
Public intColumn_Headers_Row As Integer      ' Row with Report_Name or #Column_Headers# keyword in Column A.
                                             '  - It is assumed that report data begins in the next row.
Public intReport_Name_Column As Integer      ' Column with Report_Name values
Public intTemplate_Name_Column As Integer    ' Column with Template_Name values
Public intRange_Name_Column As Integer       ' Column with Range_Name values
Public intPage_Number_Column As Integer      ' Column with Page_Number values
Public intColumns_Column As Integer          ' Column with Columns values
Public intData01_Column As Integer           ' Column with Data01 values
Public intiUnderline_Column As Integer       ' Column with iUnderline values
Public intbBold_Column As Integer            ' Column with bBold values
Public intbItalic_Column As Integer          ' Column with bItalic values
Public intsRowType_Column As Integer         ' Column with sRowType values
Public inthDetail_Column As Integer          ' Column with hDetail values
Public inthTotalInto_Column As Integer       ' Column with hTotalInto values
Public intiFactor_Column As Integer          ' Column with iFactor values
Public intCSVActiveRows As Long              ' Rows of active data in CSV tab. Value set by UniqueReports procedure.
'

Public Sub BuildReport(Optional strXLS As String)

    ' This procedure handles the creation of standard Launch Excel reports
    '  - Modify this procedure for specific reporting needs OR
    '    call custom code from LE_Custom module as necessary to avoid changing the standard
    
    ' strXLS was added to make it easier to test without using the Auto_Open procedure
    
    ' 2009-05-28 JPD Added more strLastGoodStatement
    ' 2009-07-30 JPD Enhanced to support crosstabs
    '                 - Added intPage_Number_Column
    ' 2009-07-31 JPD Added logic to handle updating existing report worksheet rather than creating new reports from templates
    ' 2009-08-04 JPD Added Call Custom_Before(strXLS) and Call Custom_After(strXLS)
    ' 2009-09-25 JPD Added step to remove |^| placeholders and replace with commas
    ' 2009-10-10 JPD added intPageNumber < 2. We only want to insert rows once.
    ' 2009-11-05 JPD Added logic to handle linked formulas
    '                Enhanced handling of NewWorkbook and SaveAs
    '                Added support for #StaticContent#
    ' 2009-11-18 JPD Added support for saving reports sent via Conductor
    ' 2009-11-25 JPD Changed logic under boolNewWorkbook from Copy to Move.
    '                Excel 2003 was not able to preserve formulas pointing to other sheets or named ranges on the same sheet
    ' 2009-12-08 JPD Changed datatypes from Integer to Long to support larger data sets
    ' 2010-01-26 JPD Added calls to Unprotect_Workbook and Protect_Workbook
    '                Added logic to allow some columns in the CSV to be omitted without affecting the VBA
    ' 2010-02-02 JPD Added support for empty Report_Name and Range_Name columns.
    ' 2010-02-23 JPD Changed location of Call Protect_Workbook and adding comments. This was not getting saved in Conductor mode.
    ' 2010-03-01 JPD Tweaked logic that unmerges cells in destination worksheet prior to pasting CSV data
    '                Updated logic for Removing the #Text# Placeholders. Now we replace with a single quote.
    ' 2010-03-02 JPD Tweaked logic for Data_Formulas intiFactor_Column
    '                Further restricted special formatting to only be copied when the correct Range Name is present
    ' 2010-03-07 JPD Set keywords boolNewWorkbook = False and boolSaveAs = False when myMode = "Conductor" And myOutput = "Printer"
    ' 2010-04-07 JPD Added call to FormatCells_AfterRows
    ' 2010-06-03 JPD Added step to remove |+| placeholders and replace with double quotes. Conductor seems to double up the double quotes
    ' 2010-10-19 JPD Updated for Excel 2010 compatibility. Excel 2010 does not always name the copied sheet the same as Excel 2007 and earlier did.
    '                Therefore, we will copy it to the end, rename it and then move it.
    ' 2010-12-01 JPD Added error logging to If WorksheetExists(strXLS, strTemplate, True)
    '                Added TextFile function
    ' 2011-02-03 JPD Fixed TextFile function
    '                Updated UniqueReports to test for empty row at bottom of CSV
    '                Changed name of Reports worksheet to LEFReports to standardize naming of worksheets LEF creates
    '                Added support for Titles
    ' 2011-04-14 JPD Added Application.Calculation = xlCalculationAutomatic to recalculate the workbook prior to saving or printing
    ' 2011-04-14 JPD Enhanced BuiltinDocumentProperties("Comments").Value to include "Excel Version " + Application.Version
    ' 2011-05-13 JPD Moved replacement of |^| and |+| Placeholders to the start of this procedure
    '                 - Replacement is now done on data in CSV tab *before* it is copied to reports
    '                 - This allows embedded commas and quotes in the names of the report and template tabs
    '                Added GetTitles to make it easier to display titles, including filter selections, in the report tabs
    '                Added HyperlinkCreate to allow for drill down capabilities
    '                Added PictureInsert to allow for importing logos or other pictures from the web or local drive
    ' 2011-05-14 JPD Added #FCFinal so that #FCFinal_SheetProtect# can be run after TextFile and PictureInsert
    ' 2011-05-29 JPD Added WkbNew.Colors = Application.Workbooks(strXLS).Colors
    '                 - This is required for custom colors in Excel 2003
    ' 2011-08-15 JPD Moved #FCProtect_ and #FCFinal_ to after RemoveExternalLinks
    ' 2011-09-30 JPD Added intCSVActiveRows
    '                Added IsNumeric test of intColumns to avoid TypeMismatch when the last named range is a single row of data
    ' 2012-01-07 JPD Moved "Copy any errors to the new workbook" before "Remove empty worksheets" to allow for No Data situations
    ' 2012-10-12 JPD Changed data type for intCSVActiveRows from Integer to Long to support CSV data > 32,768 rows
    ' 2013-10-06 JPD Added support for PageHeaderFooter procedure
        
    On Error GoTo ErrorHandler

    Dim Wkb As Workbook
    Dim Wks As Worksheet
    Dim WkbNew As Workbook
    Dim WksNew As Worksheet
    Dim strReport As String
    Dim strTemplate As String
    Dim strRangeName As String
    Dim intPageNumber As Long
    Dim intColumns As Long
    Dim intCurrentRow As Long
    Dim iOffsetColumns As Long
    Dim i As Long
    Dim intRowCount As Long
    Dim rngSource As Range
    Dim strErrorMessage As String
    Dim strSaveAsName As String
    Dim mbResult As VbMsgBoxResult
    Dim boolSaved As Boolean
    Dim strTitle As String
    Dim strFilterCaption As String
    Dim strFilterValue As String
    Dim strFilterAll As String
    Dim strFilter1Row As String
    Dim strTitlesAll As String
    Dim strTitlesAllExcept1 As String
    
    If strXLS = "" Then strXLS = myXLS
    
    ' Change |^| placeholders to commas
    Application.Workbooks(strXLS).Sheets("CSV").UsedRange.Replace What:="|^|", LookAt:=xlPart, Replacement:=","
            
    strLastGoodStatement = "Remove the |^| Placeholders(" + strXLS + "," + "CSV" + ")"
    
    ' Change |+| placeholders to double quotes
    Application.Workbooks(strXLS).Sheets("CSV").UsedRange.Replace What:="|+|", LookAt:=xlPart, Replacement:=""""
    
    strLastGoodStatement = "Remove the |+| Placeholders(" + strXLS + "," + "CSV" + ")"
    
    ' Create worksheet to store names of Report and Template worksheets used in this report
    With Application.Workbooks(strXLS)
        If Not WorksheetExists(strXLS, "LEFReports") Then
            ' Add LEFReports worksheet if it doesn't exist
            .Worksheets.Add(after:=Worksheets(.Worksheets.Count)).Name = "LEFReports"
        Else
            ' Otherwise, clear any existing data
            .Worksheets("LEFReports").UsedRange.Clear
        End If
        
        strLastGoodStatement = "WorksheetExists(" + strXLS + ", ""LEFReports"")"
        
    End With
    
    ' Create worksheet to store titles and filter data
    With Application.Workbooks(strXLS)
        If Not WorksheetExists(strXLS, "LEFTitles") Then
            ' Add LEFTitles worksheet if it doesn't exist
            .Worksheets.Add(after:=Worksheets(.Worksheets.Count)).Name = "LEFTitles"
        Else
            ' Otherwise, clear any existing data
            .Worksheets("LEFTitles").UsedRange.Clear
        End If
        
        strLastGoodStatement = "WorksheetExists(" + strXLS + ", ""LEFTitles"")"
        
    End With
    
    ' Placeholder for custom code to be run before Call FindKeywords
    Call Custom_Report_FindKeywords_Before(myXLS)

    ' Find the Keywords that pinpoint the columns and rows that contain data we care about
    Call FindKeywords(strXLS)
    
    strLastGoodStatement = "Call FindKeywords"
    
    ' Conductor requires a new workbook when sending a file, but not when printing
    If myMode = "Conductor" And myOutput = "File" Then boolNewWorkbook = True
    If myMode = "Conductor" And myOutput = "Printer" Then
        boolNewWorkbook = False
        boolSaveAs = False
    End If
    
    ' Placeholder for custom code to be run after Call FindKeywords
    Call Custom_Report_FindKeywords_After(myXLS)
    
    ' Delete old CSV files, but only if the report is being processed in classic Launch Excel mode via YSI.NetClient
    If myMode = "Launch Excel" And boolDeleteOldCSV = True Then
        ' Placeholder for custom code to be run before Call DeleteOldCSV
        Call Custom_Report_DeleteOldCSV_Before(myXLS)
        
        ' Delete CSV files more than 120 minutes old
        Call DeleteOldCSV(Workbooks(strXLS).Path & "\", 120)
    
        strLastGoodStatement = "Call DeleteOldCSV"
        
        ' Placeholder for custom code to be run after Call DeleteOldCSV
        Call Custom_Report_DeleteOldCSV_After(myXLS)
        
    ' Otherwise, delete the CSV files used in this report
    ElseIf (myMode = "Launch Excel Silverlight" Or myMode = "Conductor") And boolDeleteOldCSV = True Then
        ' Placeholder for custom code to be run before Call DeleteOldCSV
        Call Custom_Report_DeleteOldCSV_Before(myXLS)
        
        ' Delete CSV file(s) used for this report
        Call DeleteMyCSV
    
        strLastGoodStatement = "Call DeleteMyCSV"
        
        ' Placeholder for custom code to be run after Call DeleteOldCSV
        Call Custom_Report_DeleteOldCSV_After(myXLS)

    End If
    
    ' Placeholder for custom code to be run before Call UniqueReports
    Call Custom_Report_UniqueReports_Before(myXLS)
    
    ' Find the names of the report and template values.
    Call UniqueReports(strXLS)

    strLastGoodStatement = "Call UniqueReports"
    
    ' Placeholder for custom code to be run after Call UniqueReports
    Call Custom_Report_UniqueReports_After(myXLS)
    
    ' Unprotect the workbook, if that option is selected
    ' If the source template has protected sheets this code should be run from Custom_Open_CopyCSV_Before
    '    in which case myPassword is unknown. strPassword would have to be hard coded
    If boolProtectSheets Then Call Unprotect_Workbook(myXLS, myPassword)
    
    ' Delete any existing report worksheets and then create all report worksheets
    intCurrentRow = 2 ' This is the first row of data
    
    For i = intCurrentRow To Sheets("LEFReports").UsedRange.Rows.Count
        ' Set the values
        With Application.Workbooks(strXLS).Worksheets("LEFReports")
            strReport = .Cells(intCurrentRow, 1).Value
            strTemplate = .Cells(intCurrentRow, 2).Value
        End With
    
        ' Template and report may be the same if we are updating an existing report worksheet
        '   rather than creating a new report based on a template.
        '   This allows for other worksheets to have live formuals that reference the report worksheet
        If strReport <> strTemplate Then
            ' Delete the report if it exists, unless the template and the report are the same
            If WorksheetExists(strXLS, strReport) Then
                Application.Workbooks(strXLS).Worksheets(strReport).Delete
                strLastGoodStatement = "Application.Workbooks(" + strXLS + ").Worksheets(" + strReport + ").Delete"
            End If
             
            ' Copy the template to the report
            If WorksheetExists(strXLS, strTemplate, True) Then
                ' 2010-10-19 JPD Updated for Excel 2010 compatibility. Excel 2010 does not always name the copied sheet the same as Excel 2007 and earlier did.
                '                Therefore, we will copy it to the end, rename it and then move it.
                Sheets(strTemplate).Copy after:=Sheets(Sheets.Count)
                ' TODO: Verify that the report name is valid length and characters. Ideally, this would be done in the SQL script.
                Sheets(Sheets.Count).Name = strReport
                Sheets(strReport).Move before:=Sheets(strTemplate)
                strLastGoodStatement = "Sheets(" + strTemplate + " (2)).Name = " + strReport
            End If
        End If
    
        intCurrentRow = intCurrentRow + 1
    Next i
    
    ' Populate LEFTitles
    ' Initialize the variable
    intCurrentRow = 0
    
    ' Loop through the CSV to find the Titles that exist prior to intColumn_Headers_Row. Titles are always in Column A.
    For i = 1 To intColumn_Headers_Row - 1
        ' Store the value of the title
        With Application.Workbooks(strXLS).Worksheets("CSV")
            strTitle = .Cells(i, 1).Value
        End With
        
        If myMode = "Conductor" And i = 1 Then
            ' Ignore the first line for Conductor files. This has special commands for Conductor.
        ElseIf Left(strTitle, 1) = "<" Then
            ' Ignore LEF Keywords
        Else
            ' Increment the counter
            intCurrentRow = intCurrentRow + 1
        
            ' Parse the title for the filter caption and filter value
            If InStr(1, strTitle, "=") > 0 Then
                strFilterCaption = Left(strTitle, InStr(1, strTitle, "=") - 1)
                strFilterValue = Right(strTitle, Len(strTitle) - InStr(1, strTitle, "="))
                If Len(strFilter1Row) = 0 Then
                    strFilter1Row = strTitle
                Else
                    strFilter1Row = strFilter1Row + " AND " + strTitle
                End If
                If Len(strFilterAll) = 0 Then
                    strFilterAll = strTitle
                Else
                    strFilterAll = strFilterAll + Chr(10) + strTitle
                End If
            Else
                strFilterCaption = ""
                strFilterValue = ""
            End If
            
            ' Set strTitlesAll
            If Len(strTitlesAll) = 0 Then
                strTitlesAll = strTitle
            Else
                strTitlesAll = strTitlesAll + Chr(10) + strTitle
            End If
            
            ' Set strTitlesAllExcept1
            If intCurrentRow = 1 Then
                strTitlesAllExcept1 = ""
            ElseIf Len(strTitlesAllExcept1) = 0 Then
                strTitlesAllExcept1 = strTitle
            Else
                strTitlesAllExcept1 = strTitlesAllExcept1 + Chr(10) + strTitle
            End If
        
            ' Set the values
            With Application.Workbooks(strXLS).Worksheets("LEFTitles")
                .Cells(intCurrentRow, 1).Value = strTitle
                .Cells(intCurrentRow, 2).Value = strFilterCaption
                .Cells(intCurrentRow, 3).Value = strFilterValue
                .Cells(1, 2).Value = strTitlesAll           ' strTitlesAll is in cell B1
                .Cells(1, 3).Value = strTitlesAllExcept1    ' strTitlesAllExcept1 is in cell C1
                .Cells(1, 4).Value = strFilterAll           ' strFilterAll is in cell D1
                .Cells(1, 5).Value = strFilter1Row          ' strFilter1Row is in cell E1
            End With
        End If
   
    Next i
    
    ' The row under the Column Headers is the first row of report data
    intCurrentRow = intColumn_Headers_Row + 1
    
    ' Set the first values
    intRowCount = 0
    iOffsetColumns = 0
    
    ' Start looping through the data to find how many rows are in each Range_Name
    For i = intCurrentRow To intCSVActiveRows ' 2011-09-30 JPD Changed from Sheets("CSV").UsedRange.Rows.Count
        With Application.Workbooks(strXLS).Worksheets("CSV")
            strReport = .Cells(intCurrentRow, intReport_Name_Column).Value
            strTemplate = .Cells(intCurrentRow, intTemplate_Name_Column).Value
            strRangeName = .Cells(intCurrentRow, intRange_Name_Column).Value
            intPageNumber = .Cells(intCurrentRow, intPage_Number_Column).Value
            intColumns = .Cells(intCurrentRow, intColumns_Column).Value
            ' Reset iOffsetColumns if this is the first page of a range
            If intPageNumber < 2 Then iOffsetColumns = 0
        End With
        
        If strRangeName = Sheets("CSV").Cells(intCurrentRow + intRowCount + 1, intRange_Name_Column).Value And _
              strReport = Sheets("CSV").Cells(intCurrentRow + intRowCount + 1, intReport_Name_Column).Value And _
              intPageNumber = Sheets("CSV").Cells(intCurrentRow + intRowCount + 1, intPage_Number_Column).Value Then
            ' Increment the row count
            intRowCount = intRowCount + 1
        Else
            ' Copy the range from CSV to the report
            If strRangeName = "#StaticContent#" Then
                ' #StaticContent# is a placeholder in the CSV file for tabs that have no real CSV data.
                ' This creates a row in the LEFReports tab so that if boolNewWorkbook=True the static
                '   tabs get transfered to Book1
            ElseIf Trim(strRangeName) = "" Then
                ' If the range name is empty then there is nothing to do
            ElseIf NamedRangeExists(Sheets(strReport), strRangeName, True) Then
                ' Test to see if this range grows vertically, based on the number of rows of data
                ' 2009-10-10 JPD added intPageNumber < 2. We only want to insert rows once.
                If Left(strRangeName, 2) = "v_" And intPageNumber < 2 Then
                    Call DynamicVertical(strXLS, strReport, strRangeName, intRowCount + 1)
                End If
                
                With Worksheets("CSV")
                    ' Be sure the destination has no merged cells. PasteSpecial can't handle merged cells.
                    Worksheets(strReport).Range(Worksheets(strReport).Cells( _
                        Worksheets(strReport).Range(strRangeName).Row, _
                          Worksheets(strReport).Range(strRangeName).Column + iOffsetColumns), _
                        Worksheets(strReport).Cells(Worksheets(strReport).Range(strRangeName).Row + intRowCount, _
                          Worksheets(strReport).Range(strRangeName).Column + iOffsetColumns + intColumns - 1)).MergeCells = False
                        
                    ' Copy data for this range from the CSV tab
                    .Range(.Cells(intCurrentRow, intData01_Column), _
                        .Cells(intCurrentRow + intRowCount, intData01_Column + intColumns - 1)).Copy
                
                    ' Paste the data into the destination
                    Worksheets(strReport).Cells(Worksheets(strReport).Range(strRangeName).Row, _
                        Worksheets(strReport).Range(strRangeName).Column + iOffsetColumns).PasteSpecial Paste:=xlPasteValues
                    strLastGoodStatement = "Worksheets(" + strReport + ").Range(" + strRangeName + ").PasteSpecial Paste:=xlPasteValues"
                
                    ' Test to see if this range is using the formatting columns
                    ' First check to see that there are valid columns for all three formatting columns
                    If intiUnderline_Column > 0 And intbBold_Column = intiUnderline_Column + 1 And _
                       intbItalic_Column = intiUnderline_Column + 2 Then
                        If Len(Worksheets("CSV").Cells(intCurrentRow, intiUnderline_Column).Value) > 0 _
                            And (strRangeName = "Data" Or strRangeName = "Data_Formulas") Then ' This is hard coded to only support these strRangeName
                            ' If it is then copy the formatting
                            .Range(.Cells(intCurrentRow, intiUnderline_Column), _
                                .Cells(intCurrentRow + intRowCount, intiUnderline_Column + 2)).Copy
                            
                            ' Paste it on the far right side of the range
                            Worksheets(strReport).Cells(Worksheets(strReport).Range(strRangeName).Row, _
                                Worksheets(strReport).Range(strRangeName).Column + (iOffsetColumns + intColumns)).PasteSpecial Paste:=xlPasteValues
                        End If
                    End If
                    
                    ' Test to see if this range is using the live formula columns
                    ' First check to see that there are valid columns for all three formatting columns and all four
                    '   live formula columns. They must be contiguous.
                    If intiUnderline_Column > 0 And intbBold_Column = intiUnderline_Column + 1 And _
                       intbItalic_Column = intiUnderline_Column + 2 And intsRowType_Column = intiUnderline_Column + 3 And _
                       inthDetail_Column = intiUnderline_Column + 4 And inthTotalInto_Column = intiUnderline_Column + 5 And _
                       intiFactor_Column = intiUnderline_Column + 6 Then
                        If Len(Worksheets("CSV").Cells(intCurrentRow, intsRowType_Column).Value) > 0 _
                            And strRangeName = "Data_Formulas" Then ' This is hard coded to only support this strRangeName
                            ' If it is then copy the formatting and the formula columns
                            .Range(.Cells(intCurrentRow, intiUnderline_Column), _
                                .Cells(intCurrentRow + intRowCount, intiUnderline_Column + 6)).Copy

                            ' Paste it on the far right side of the range
                            Worksheets(strReport).Cells(Worksheets(strReport).Range(strRangeName).Row, _
                                Worksheets(strReport).Range(strRangeName).Column + (iOffsetColumns + intColumns)).PasteSpecial Paste:=xlPasteValues
                        End If
                    End If
                End With
                
                ' Increment iOffsetColumns (This will be reset to 0 if we start a new range)
                iOffsetColumns = iOffsetColumns + intColumns
                
            End If
            
            ' Set the new strRangeName and intColumns
            strRangeName = Sheets("CSV").Cells(intCurrentRow + intRowCount + 1, intRange_Name_Column).Value
            ' 2011-09-30 JPD Added IsNumeric test of intColumns to avoid TypeMismatch when the last named range is a single row of data
            If IsNumeric(Sheets("CSV").Cells(intCurrentRow + intRowCount + 1, intColumns_Column).Value) Then
                intColumns = Sheets("CSV").Cells(intCurrentRow + intRowCount + 1, intColumns_Column).Value
            Else
                intColumns = 1
            End If
            
            'Set the new intCurrentRow
            intCurrentRow = intCurrentRow + intRowCount + 1
            
            ' Reset intRowCount
            intRowCount = 0
        
            ' The last section that was copied remains selected. Change that to the first cell.
            If Not Trim(strReport) = "" Then
                Sheets(strReport).Activate
                Sheets(strReport).Cells(1, 1).Select
            End If
        
        End If
    
    Next i
    
    ' Custom procedures to be run before standard post processing
    Call Custom_Report_Processing_Before(strXLS)

    ' Post processing of each report
    ' 1 - Change |^| placeholders to commas  *** 2011-05-13 JPD This was moved to the start of this procedure and is now done on the CSV tab
    ' 1.1 - Change |+| placeholders to double quotes *** 2011-05-13 JPD This was moved to the start of this procedure and is now done on the CSV tab
    ' 2 - Remove #Text# Placeholders
    ' 2.1 FormatCells that have keyword like "#Filter"
    ' 3 - FormatCells that have keyword like "#FC_"
    ' 4 - Build formulas for account tree total rows
    ' 5 - BuildFormulaR1C1
    ' 6 - Format the data rows (Bold, Italic, Borders)
    '      - Also deletes the columns with the formatting data
    
    intCurrentRow = 2 ' This is the first row of data
    
    For i = intCurrentRow To Application.Workbooks(strXLS).Worksheets("LEFReports").UsedRange.Rows.Count
        ' Set the values
        With Application.Workbooks(strXLS).Worksheets("LEFReports")
            strReport = .Cells(intCurrentRow, 1).Value
        End With
    
        If WorksheetExists(strXLS, strReport) Then
            ' Remove the |^| Placeholders *** 2011-05-13 JPD This was moved to the start of this procedure and is now done on the CSV tab
            'Application.Workbooks(strXLS).Sheets(strReport).UsedRange.Replace What:="|^|", LookAt:=xlPart, Replacement:=","
            
            'strLastGoodStatement = "Remove the |^| Placeholders(" + strXLS + "," + strReport + ")"
            
            ' Remove the |+| Placeholders *** 2011-05-13 JPD This was moved to the start of this procedure and is now done on the CSV tab
            'Application.Workbooks(strXLS).Sheets(strReport).UsedRange.Replace What:="|+|", LookAt:=xlPart, Replacement:=""""
            
            'strLastGoodStatement = "Remove the |+| Placeholders(" + strXLS + "," + strReport + ")"
            
            ' Remove the #Text# Placeholders
            Application.Workbooks(strXLS).Sheets(strReport).UsedRange.Replace What:="#Text#", LookAt:=xlPart, Replacement:="'"
            
            strLastGoodStatement = "Remove the #Text# Placeholders(" + strXLS + "," + strReport + ")"
            
            ' 2010-05-13 JPD Find and replace titles
            Call GetTitles(strXLS, strReport)
            strLastGoodStatement = "GetTitles(" + strXLS + "," + strReport + ")"
            
            ' 2010-05-13 JPD Create hyperlinks
            Call HyperlinkCreate(strXLS, strReport, "#Hyperlink_")
            strLastGoodStatement = "HyperlinkCreate #Hyperlink_(" + strXLS + "," + strReport + ")"
            
            ' 2010-04-28 JPD Format cells based on hard coded logic OR the format of named ranges in the "Format" tab
            Call FormatCells(strXLS, strReport, "#FC_", "Format")
            strLastGoodStatement = "FormatCells #FC_(" + strXLS + "," + strReport + ")"
            
            ' 2013-10-06 JPD Added PageHeaderFooter
            Call PageHeaderFooter(strXLS, strReport, "")
            strLastGoodStatement = "PageHeaderFooter (" + strXLS + "," + strReport + ")"
            
            ' Build formulas for account tree total rows
            ' 2009-08-02 JPD At present this is hard coded to support named range Data_Formulas
            Call BuildTotalRows(strXLS, strReport, "Data_Formulas")
            
            strLastGoodStatement = "BuildTotalRows(" + strXLS + "," + strReport + ")"
            
            ' Build formulas for account tree total columns or any other properly formatted FormulaR1C1
            Call BuildFormulaR1C1(strXLS, strReport)
            
            strLastGoodStatement = "BuildFormulaR1C1(" + strXLS + "," + strReport + ")"
            
            ' Format rows.
            ' 2009-08-02 JPD At present this is hard coded to support named ranges Data and Data_Formulas
            Call FormatRows(strXLS, strReport, "Data")
            Call FormatRows(strXLS, strReport, "Data_Formulas")
            
            strLastGoodStatement = "FormatRows(" + strXLS + "," + strReport + ")"

            ' 2010-04-28 JPD Format cells after the rows have been formatted. Allows for overriding row level underline/bold/italic
            '                Format cells based on hard coded logic OR the format of named ranges in the "Format" tab
            Call FormatCells(strXLS, strReport, "#FCAfter_", "Format")
            strLastGoodStatement = "FormatCells #FCAfter_(" + strXLS + "," + strReport + ")"
        
            ' 2013-10-06 JPD Added PageHeaderFooterAfter
            Call PageHeaderFooter(strXLS, strReport, "After")
            strLastGoodStatement = "PageHeaderFooterAfter (" + strXLS + "," + strReport + ")"
        
            ' 2010-12-01 JPD Create text files
            Call TextFile(strXLS, strReport, "#TextFile_", "Format")
            strLastGoodStatement = "TextFile #TextFile_(" + strXLS + "," + strReport + ")"
        
            ' 2011-05-13 JPD Insert pictures last, after all column and row formatting has been done
            Call PictureInsert(strXLS, strReport, "#Picture_")
            strLastGoodStatement = "PictureInsert #Picture_(" + strXLS + "," + strReport + ")"
        
            ' 2011-08-15 JPD #FCFinal_ moved to after RemoveExternalLinks
        End If
         
        intCurrentRow = intCurrentRow + 1
    Next i
   
    ' Custom procedures to be run after standard post processing
    ' 2010-04-28 JPD Moved this call so that the processing happens before new workbook is created
    Call Custom_Report_Processing_After(strXLS)
   
   ' Create a new workbook and copy over the report worksheets
   If boolNewWorkbook Then
   
        ' Create a new workbook
        Set WkbNew = Workbooks.Add
        strLastGoodStatement = "WkbNew = Workbooks.Add"
        
        ' Copy the colors. This is required for custom colors in Excel 2003
        WkbNew.Colors = Application.Workbooks(strXLS).Colors
        strLastGoodStatement = "WkbNew.Colors = Application.Workbooks(strXLS).Colors"
        
        'Copy the CSV, if desired
        If boolDisplayCSV Then
            Application.Workbooks(strXLS).Sheets("CSV").Copy after:=WkbNew.Worksheets("Sheet1")
        End If
        
        ' Copy the reports to a new workbook
            intCurrentRow = 2 ' This is the first row of data
        
        For i = intCurrentRow To Application.Workbooks(strXLS).Worksheets("LEFReports").UsedRange.Rows.Count
            ' Set the values
            With Application.Workbooks(strXLS).Worksheets("LEFReports")
                strReport = .Cells(intCurrentRow, 1).Value
            End With
        
            If WorksheetExists(strXLS, strReport) Then
                ' 2009-11-25 JPD Split the logic for Copy or Move depending on version of Excel.
                '                Excel 2003 does not preserve formulas pointing to other sheets when copied
                '                Excel 2007 does not preserve Print_Titles when moved or copied
                If Application.Version >= 12 Then ' Excel 2007
                    ' Move the report to the new workbook
                    Application.Workbooks(strXLS).Sheets(strReport).Copy before:=WkbNew.Worksheets("Sheet1")
                    strLastGoodStatement = "Application.Workbooks(" + strXLS + ").Sheets(" + strReport + ").Copy before:=WkbNew.Worksheets(""Sheet1"")"
                    ' Reset Print_Titles. For some reason it gets confused when moved or copied to a new workbook in Excel 2007
                   If NamedRangeExists(Sheets(strReport), "Print_Titles", False) Then
                        WkbNew.Worksheets(strReport).Names("Print_Titles").RefersTo = Workbooks(strXLS).Worksheets(strReport).Names("Print_Titles").RefersTo
                    End If
                    strLastGoodStatement = "WkbNew.Worksheets(" + strReport + ").Names(""Print_Titles"").RefersTo = ""="" + strReport + ""!$1:$6"""
                Else ' Prior to Excel 2007
                    ' Move the report to the new workbook
                    Application.Workbooks(strXLS).Sheets(strReport).Move before:=WkbNew.Worksheets("Sheet1")
                    strLastGoodStatement = "Application.Workbooks(" + strXLS + ").Sheets(" + strReport + ").Move before:=WkbNew.Worksheets(""Sheet1"")"
                
                End If
                
            End If
             
            intCurrentRow = intCurrentRow + 1
        Next i
           
        ' Remove links in formulas that point to the template
        RemoveExternalLinks (WkbNew.Name)
        
        ' Copy any errors to the new workbook
        If WorksheetExists(strXLS, "Errors") Then
            Application.Workbooks(strXLS).Sheets("Errors").Copy before:=WkbNew.Worksheets(1)
            strLastGoodStatement = "Application.Workbooks(" + strXLS + ").Sheets(""Errors"").Copy before:=WkbNew.Worksheets(1)"
        End If
       
       ' Remove empty worksheets
        For Each WksNew In WkbNew.Worksheets
            If Left(WksNew.Name, 5) = "Sheet" Then
                WksNew.Delete
                strLastGoodStatement = "WksNew.Delete"
            End If
        Next WksNew
        
        ' Activate the first worksheet in the new book. It will be the errors or the first report
        WkbNew.Worksheets(1).Activate
        
        ' We want to close this "source" workbook after Auto_Open has completed its cleanup
        boolCloseMe = True
        
        ' From now on we will be working with the new workbook, rather than the source
        strXLS = WkbNew.Name  ' 2009-11-05 JPD Added
        myXLS = WkbNew.Name  ' 2009-11-05 JPD Added

    Else ' If boolNewWorkbook
        ' Clean up this workbook
        If Not myMode = "Conductor" And myOutput = "Printer" Then
            ' We need the LEFReports sheet when printing to determine which sheets to print
            If WorksheetExists(strXLS, "LEFReports") Then Application.Workbooks(strXLS).Sheets("LEFReports").Delete
        End If
        
        If WorksheetExists(strXLS, "Message") Then Application.Workbooks(strXLS).Sheets("Message").Visible = False
        If Not boolDisplayCSV Then Application.Workbooks(strXLS).Sheets("CSV").Delete
        If WorksheetExists(strXLS, "LEFTitles") Then Application.Workbooks(strXLS).Sheets("LEFTitles").Delete
        
    End If ' If boolNewWorkbook
    
    ' Protect the workbook, if that option is selected
    If boolProtectSheets Then Call Protect_Workbook(myXLS, myPassword)
    
    ' 2011-08-15 JPD Added #FCProtect_ so that #FCFinal_ can still be run last, if needed, to set the focus on a particular cell
    For Each Wks In Application.Workbooks(strXLS).Worksheets
        ' Typically, only #FCProtect_SheetProtect# would be called here
        Call FormatCells(strXLS, Wks.Name, "#FCProtect_", "Format")
        strLastGoodStatement = "FormatCells #FCProtect_(" + strXLS + "," + Wks.Name + ")"
    Next Wks
    
    For Each Wks In Application.Workbooks(strXLS).Worksheets
        ' 2013-10-06 JPD Added PageHeaderFooterProtect
        Call PageHeaderFooter(strXLS, Wks.Name, "Protect")
        strLastGoodStatement = "PageHeaderFooterProtect (" + strXLS + "," + strReport + ")"
    Next Wks
    
    ' 2011-08-15 JPD Moved #FCFinal_ to after RemoveExternalLinks
    For Each Wks In Application.Workbooks(strXLS).Worksheets
        ' Typically, only #FCFinal_SheetProtect# would need to be run this late
        Call FormatCells(strXLS, Wks.Name, "#FCFinal_", "Format")
        strLastGoodStatement = "FormatCells #FCFinal_(" + strXLS + "," + Wks.Name + ")"
    Next Wks
    
    For Each Wks In Application.Workbooks(strXLS).Worksheets
        ' 2013-10-06 JPD Added PageHeaderFooterFinal
        Call PageHeaderFooter(strXLS, Wks.Name, "Final")
        strLastGoodStatement = "PageHeaderFooterProtect (" + strXLS + "," + strReport + ")"
    Next Wks
    
    ' Add a comment so we know when the workbook was created and which version of the framework was used
    Workbooks(myXLS).BuiltinDocumentProperties("Comments").Value = "Created " & Now & _
        " by Launch Excel Framework Version " & conLEFVersion & " with Excel Version " + Application.Version
    
    ' 2011-04-14 JPD Recalculate the workbook
    Application.Calculation = xlCalculationAutomatic
    
    If myMode = "Conductor" And myOutput = "File" Then
        boolSaved = SaveWorkbook(strXLS, myOutputFile, False, True)
        strXLS = myXLS

    ElseIf boolSaveAs Then
        strSaveAsName = Application.GetSaveAsFilename
        If strSaveAsName = "False" Then
            ' Warn the user and provide another opportunity to save
            If Not boolNewWorkbook Then
                mbResult = MsgBox("This workbook will be overwritten the next time this report " & _
                    "is run from Voyager unless it is saved to another location. Do you want to save it now?", vbYesNo, "Workbook will be overwritten.")
                If mbResult = vbYes Then
                    strSaveAsName = Application.GetSaveAsFilename
                Else
                    MsgBox "Save As was canceled by user", vbOKOnly, "Canceled"
                End If
            End If
        End If
        
        If strSaveAsName <> "False" Then
            ' Make sure the name is properly formatted
            If LCase(Right(strSaveAsName, 1)) = "." Then strSaveAsName = strSaveAsName + "xls"
            If LCase(Right(strSaveAsName, 4)) <> ".xls" Then strSaveAsName = strSaveAsName + ".xls"
            
            ' Turn on the alerts so we get a warning before overwriting
            Application.DisplayAlerts = True
            
            boolSaved = SaveWorkbook(strXLS, strSaveAsName, False, True)
            strXLS = myXLS  ' 2009-09-04 JPD Added
            
            ' Turn off the alerts
            Application.DisplayAlerts = False
            
            If Not boolSaved Then
                MsgBox "Save As was not successful", vbOKOnly, "File not saved"
            End If
            
        End If
        
    End If
    


ExitProcedure:
    Set WkbNew = Nothing
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("BuildReport", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "BuildReport"
            #End If
            
    End Select
    Resume ExitProcedure
    

End Sub

Public Sub FindKeywords(Optional strXLS As String)
    ' 2009-02-15 JPD    Created
    ' 2009-07-31 JPD    Added Keywords: <NewWorkbook>, <SaveAs>, <DisplayCSV>, <DeleteOldCSV>
    ' 2009-08-02 JPD    Added Column Headers: sRowType, hDetail, hTotalInto, iFactor
    ' 2010-01-26 JPD    Changed from #Column_Headers# to Report_Name to save a column in the CSV file
    '                   Page_Number column is now optional.
    '                   sRowType column is now optional.
    '                   Added Keywords: <ProtectSheets>, <Password>
    ' 2010-01-27 JPD    Changed <SaveAs> default to No
    ' 2010-03-02 JPD    Fixed bug in logic for intsRowType_Column
    ' 2011-09-30 JPD    Voyager 7.0 has a leading space on the column names, so check for " Report_Name"
    '                   Thanks to Dave Chapman for finding this nuance and supplying the code
    ' 2011-10-04 JPD    Change placement of removal of leading spaces on the column names to properly
    '                   support use of #Column_Headers# as the first column
    
    ' This procedure populates the global variables for column and row location of Keywords
    
    Dim i As Long
    
    If strXLS = "" Then strXLS = myXLS
    
    ' Keyword <NewWorkbook=No>. If missing, default to true
    If FirstMatch("<NewWorkbook=No>", strXLS, "CSV", "A:A") = -1 Then
        boolNewWorkbook = True
    Else
        boolNewWorkbook = False
    End If
    
    ' Keyword <SaveAs=Yes>. If missing, default to false
    If FirstMatch("<SaveAs=Yes>", strXLS, "CSV", "A:A") = -1 Then
        boolSaveAs = False
    Else
        boolSaveAs = True
    End If
    
    ' Keyword <DisplayCSV=Yes>. If missing, default to false
    If FirstMatch("<DisplayCSV=Yes>", strXLS, "CSV", "A:A") = -1 Then
        boolDisplayCSV = False
    Else
        boolDisplayCSV = True
    End If
    
    ' Keyword <ProtectSheets=Yes>. If missing, default to false
    If FirstMatch("<ProtectSheets=Yes>", strXLS, "CSV", "A:A") = -1 Then
        boolProtectSheets = False
    Else
        boolProtectSheets = True
    End If
    
    ' Keyword <Password=>. If missing, default to ""
    If FirstMatch("<Password=*", strXLS, "CSV", "A:A") = -1 Then
        myPassword = ""
    Else
        ' Get the complete value of the cell
        myPassword = Worksheets("CSV").Cells(FirstMatch("<Password=*", strXLS, "CSV", "A:A"), 1).Value
        ' Parse it to find the password. #Blank# is a keywork meaning an empty password
        If myPassword = "<Password=#Blank#>" Then
            myPassword = ""
        Else
            myPassword = Mid(myPassword, 11, Len(myPassword) - 11)
        End If
    End If
    
    ' Keyword <DeleteOldCSV=No>. If missing, default to true
    If FirstMatch("<DeleteOldCSV=No>", strXLS, "CSV", "A:A") = -1 Then
        boolDeleteOldCSV = True
    Else
        boolDeleteOldCSV = False
    End If
    
    ' First find the row with the column headers. This will always be in Column A
    ' 2010-01-26 JPD Changed from #Column_Headers# to Report_Name to save a column in the CSV file
    intColumn_Headers_Row = FirstMatch("Report_Name", strXLS, "CSV", "A:A")
    
    ' 2011-09-30 JPD Voyager 7.0 has a leading space on the column names, so check for " Report_Name"
    '                Thanks to Dave Chapman for finding this nuance and supplying the code
    If intColumn_Headers_Row = -1 Then
        intColumn_Headers_Row = FirstMatch(" Report_Name", strXLS, "CSV", "A:A")
    End If
    
    ' 2010-01-26 JPD Check for #Column_Headers# for backward compatibility with prior CSV formats
    If intColumn_Headers_Row = -1 Then
        intColumn_Headers_Row = FirstMatch("#Column_Headers#", strXLS, "CSV", "A:A")
    End If

    ' 2011-10-04 JPD Voyager 7.0 has a leading space on the column names, so check for " #Column_Headers#"
    If intColumn_Headers_Row = -1 Then
        intColumn_Headers_Row = FirstMatch(" #Column_Headers#", strXLS, "CSV", "A:A")
    End If

    ' 2011-10-04 JPD Voyager 7.0 has a leading space on the column names, so remove all leading spaces in this row
    If intColumn_Headers_Row > 0 Then
        With Application.Workbooks(strXLS).Sheets("CSV")
            For i = 1 To .UsedRange.Columns.Count
                .Cells(intColumn_Headers_Row, i).Value = Trim(.Cells(intColumn_Headers_Row, i).Value)
            Next i
        End With
    End If
    
    ' Then find the columns with Keywords for Column Headers
    intReport_Name_Column = FirstMatch("Report_Name", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    intTemplate_Name_Column = FirstMatch("Template_Name", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    intRange_Name_Column = FirstMatch("Range_Name", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    
    ' 2010-01-26 JPD Page_Number column is now optional.
    '                Logic still requires a valid column so we will set it outside the used range
    If FirstMatch("Page_Number", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row) = -1 Then
        intPage_Number_Column = Workbooks(strXLS).Sheets("CSV").UsedRange.Columns.Count + 1
    Else
        intPage_Number_Column = FirstMatch("Page_Number", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    End If

    intColumns_Column = FirstMatch("Columns", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    intData01_Column = FirstMatch("Data01", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    intiUnderline_Column = FirstMatch("iUnderline", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    intbBold_Column = FirstMatch("bBold", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    intbItalic_Column = FirstMatch("bItalic", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    
    ' 2010-01-26 JPD sRowType column is now optional.
    ' 2010-03-02 JPD Fixed bug in logic.
    '                Logic still requires a valid column so we will set it outside the used range
    If FirstMatch("sRowType", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row) = -1 Then
        intsRowType_Column = Workbooks(strXLS).Sheets("CSV").UsedRange.Columns.Count + 1
    Else
        intsRowType_Column = FirstMatch("sRowType", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    End If
    
    inthDetail_Column = FirstMatch("hDetail", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    inthTotalInto_Column = FirstMatch("hTotalInto", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    intiFactor_Column = FirstMatch("iFactor", strXLS, "CSV", intColumn_Headers_Row & ":" & intColumn_Headers_Row)
    
End Sub

Public Sub UniqueReports(Optional strXLS As String)
    ' 2009-03-10 JPD    Created
    ' 2010-01-27 JPD    Updated to handle Grand Total in Report_Name column now that it is column A
    ' 2010-03-03 JPD    Updated. Conductor seems to add several periods at the end of column A of the CSV file.
    '                   They need to be removed.
    ' 2011-02-03 JPD    Updated. Conductor seems to add an empty row to the end of the CSV file.
    '                   It needs to be removed.
    ' 2011-09-20 JPD    Added intRows to eliminate error if CSV has <= 5
    ' 2011-09-30 JPD    Added intCSVActiveRows to prevent empty rows from being included in unique reports
    ' 2011-10-17 JPD    Improved support for empty column A which could happen if #Column_Headers# keyword is in Column A.
    ' 2012-01-07 JPD    Log error if there is no data in the CSV file
    
    ' This procedure creates a unique list of Report/Template combinations
    Dim intCount As Long
    Dim intRows As Long
    
    If strXLS = "" Then strXLS = myXLS
    
        With Application.Workbooks(strXLS).Worksheets("CSV")
            ' Find the last row of data
            intRows = .UsedRange.Rows.Count
            intCSVActiveRows = .UsedRange.Rows.Count
            ' Loop up backwards looking for stray data at the end of the CSV range
            ' at least 3 times (0,1,2) in case there is a
            ' space in the last row or two of the named range.
            For intCount = 0 To 2
                ' 2010-01-27 If the last cell is Grand Total we need to delete it so it's not one of the report names
                ' 2010-03-03 If the last cell is a period we need to delete it so it's not one of the report names
                ' 2011-02-03 If the last cell is empty we need to delete it so it's not one of the report names
                ' 2011-10-17 JPD Add check of both columns A and B = "" in case ColumnHeaders is still the first column
                If .Cells(intRows - intCount, 1).Value = "Grand Total" Or _
                        .Cells(intRows - intCount, 1).Value = "." Or _
                        (Trim(.Cells(intRows - intCount, 1).Value) = "" And Trim(.Cells(intRows - intCount, 2).Value) = "") Then
                    .Cells(intRows - intCount, 1).Clear
                    ' Decrement intCSVActiveRows
                    intCSVActiveRows = intCSVActiveRows - 1
                End If
            Next intCount
            
            ' 2012-01-07 Test to see if there are any reports
            If intCSVActiveRows > intColumn_Headers_Row Then
                .Range(.Cells(intColumn_Headers_Row, intReport_Name_Column), _
                .Cells(intCSVActiveRows, intTemplate_Name_Column)).AdvancedFilter _
                Action:=xlFilterCopy, CopyToRange:= _
                Application.Workbooks(strXLS).Worksheets("LEFReports").Cells(1, 1), Unique:=True
            Else
                ' 2012-01-07 If no reports, then log the error
                Call LogError("UniqueReports", "There is no data for these filter values")
            End If
        End With
        
End Sub

Public Sub FormatRows(strXLS As String, strReport As String, strRange As String)
    ' 2009-05-19 JPD Created
    ' 2009-05-26 JPD Removed .TintAndShade = 0. This is not supported in Excel 2003 and prior.
    ' 2009-08-02 JPD Added strRange parameter.
    ' 2009-11-20 JPD Underlining now applies to all columns
    ' 2009-12-08 JPD Changed datatypes from Integer to Long to support larger data sets
    ' 2010-02-27 JPD Added support for iUnderline = 3. This is a single underline.
    
    '                This will format the Underline, Bold and Italics for each row then delete the formatting data
                     
    Dim intCurrentRow As Long
    Dim i As Long
    Dim intColumn_Data As Long
    Dim intColumn_Underline As Long
    Dim intColumn_Bold As Long
    Dim intColumn_Italic As Long
    
    If NamedRangeExists(Sheets(strReport), strRange, False) Then
    
        intCurrentRow = Worksheets(strReport).Range(strRange).Row
        intColumn_Data = Worksheets(strReport).Range(strRange).Column
        intColumn_Underline = Worksheets(strReport).UsedRange.Columns.Count - 2
        intColumn_Bold = Worksheets(strReport).UsedRange.Columns.Count - 1
        intColumn_Italic = Worksheets(strReport).UsedRange.Columns.Count
        
        With Worksheets(strReport)
            For i = intCurrentRow To .UsedRange.Rows.Count
                strLastGoodStatement = "FormatRows(" + strXLS + "," + strReport + ") Begin intCurrentRow = " + CStr(intCurrentRow)
                ' Overline all columns
                If .Cells(intCurrentRow, intColumn_Underline).Value = 1 Then
                    With .Range(.Cells(intCurrentRow, intColumn_Data), _
                            .Cells(intCurrentRow, Worksheets(strReport).UsedRange.Columns.Count))
                        With .Borders(xlEdgeTop)
                            .ColorIndex = 0
                            .Weight = xlThin
                            .LineStyle = xlContinuous
                        End With
                    End With
                End If
    
                ' Overline and Double Underline all columns
                If .Cells(intCurrentRow, intColumn_Underline).Value = 2 Then
                    With .Range(.Cells(intCurrentRow, intColumn_Data), _
                            .Cells(intCurrentRow, Worksheets(strReport).UsedRange.Columns.Count))
                        With .Borders(xlEdgeTop)
                            .ColorIndex = 0
                            .Weight = xlThin
                            .LineStyle = xlContinuous
                        End With
                        With .Borders(xlEdgeBottom)
                            .ColorIndex = 0
                            .Weight = xlThin
                            .LineStyle = xlDouble   ' 2010-02-27 If LineStyle is before weight the line is only single
                        End With
                    End With
                End If
                
                ' Underline all columns
                If .Cells(intCurrentRow, intColumn_Underline).Value = 3 Then
                    With .Range(.Cells(intCurrentRow, intColumn_Data), _
                            .Cells(intCurrentRow, Worksheets(strReport).UsedRange.Columns.Count))
                        With .Borders(xlEdgeBottom)
                            .ColorIndex = 0
                            .Weight = xlThin
                            .LineStyle = xlContinuous
                        End With
                    End With
                End If
                
                ' Bold all columns
                If .Cells(intCurrentRow, intColumn_Bold).Value = -1 Then
                    .Range(.Cells(intCurrentRow, intColumn_Data), _
                        .Cells(intCurrentRow, Worksheets(strReport).UsedRange.Columns.Count)).Font.Bold = True
                End If
                
                ' Italics all columns
                If .Cells(intCurrentRow, intColumn_Italic).Value = -1 Then
                    .Range(.Cells(intCurrentRow, intColumn_Data), _
                        .Cells(intCurrentRow, Worksheets(strReport).UsedRange.Columns.Count)).Font.Italic = True
                End If
                
                intCurrentRow = intCurrentRow + 1
            Next i
        
        strLastGoodStatement = "FormatRows(" + strXLS + "," + strReport + ") Completed intCurrentRow = " + CStr(intCurrentRow)

        ' Delete the formatting columns
        .Columns(intColumn_Italic).Delete
        .Columns(intColumn_Bold).Delete
        .Columns(intColumn_Underline).Delete
        
        End With
    
    End If
    
End Sub

Public Sub BuildTotalRows(strXLS As String, strReport As String, strRange As String)
    ' 2009-08-02 JPD Created
    ' 2009-12-08 JPD Changed datatypes from Integer to Long to support larger data sets
    ' 2011-03-16 JPD Fixed bug by adding intRowCount = 0 after building the formula
    
    '                   This will create live Excel formulas for the total rows of an account tree
    '                   Looks for cells with a value of "#Formula-TotalRow#"
                     
    Dim intCurrentRow As Long
    Dim intRow_Detail As Long
    Dim intCurrentColumn As Long
    Dim i As Long
    Dim ii As Long
    Dim intColumn_Data As Long
    Dim intsRowType_Column As Long
    Dim inthDetail_Column As Long
    Dim inthTotalInto_Column As Long
    Dim intiFactor_Column As Long
    Dim strFormula As String
    Dim strhTotal As String
    Dim striFactor_Total As String
    Dim striFactor_Current As String
    Dim intRowCount As Long
    
    ' Find the columns and rows we need to identify which rows total into each other
    If NamedRangeExists(Sheets(strReport), strRange, False) Then
        intCurrentRow = Worksheets(strReport).Range(strRange).Row
        intColumn_Data = Worksheets(strReport).Range(strRange).Column
        intsRowType_Column = Worksheets(strReport).UsedRange.Columns.Count - 3
        inthDetail_Column = Worksheets(strReport).UsedRange.Columns.Count - 2
        inthTotalInto_Column = Worksheets(strReport).UsedRange.Columns.Count - 1
        intiFactor_Column = Worksheets(strReport).UsedRange.Columns.Count
        
        With Worksheets(strReport)
            For i = intCurrentRow To .UsedRange.Rows.Count
                strLastGoodStatement = "BuildTotalRows(" + strXLS + "," + strReport + ") Begin intCurrentRow = " + CStr(intCurrentRow)
                ' Determine if this is a Total row
                If .Cells(intCurrentRow, intsRowType_Column).Value = "T" Then
                    ' Find the first column in this row with a value of "#Formula-TotalRow#"
                    intCurrentColumn = FirstMatch("#Formula-TotalRow#", strXLS, strReport, intCurrentRow & ":" & intCurrentRow)
                    
                    'If there's a match, then build the first formula
                    If intCurrentColumn <> -1 Then
                        ' Get the variables for this total row
                        strhTotal = .Cells(intCurrentRow, inthDetail_Column).Value
                        striFactor_Total = .Cells(intCurrentRow, intiFactor_Column).Value
                        ' Set intRow_Detail as the first row of this named range
                        intRow_Detail = Worksheets(strReport).Range(strRange).Row
                        ' Reset variables
                        intRowCount = 0
                        strFormula = "="

                        ' Loop through the rows above the current total
                        For ii = intRow_Detail To intCurrentRow - 1
                            ' Get striFactor_Current for this row
                            striFactor_Current = .Cells(intRow_Detail, intiFactor_Column).Value
                            ' Test to see if this detail row totals into intCurrentRow
                            If strhTotal = .Cells(intRow_Detail, inthTotalInto_Column).Value Then
                                ' If this is part of a contiguous section then increment the row count
                                If strhTotal = .Cells(intRow_Detail + 1, inthTotalInto_Column).Value And _
                                    striFactor_Current = .Cells(intRow_Detail + 1, intiFactor_Column).Value Then
                                    ' Increment the row count
                                    intRowCount = intRowCount + 1
                                
                                'Otherwise, append this part of the formula
                                Else
                                    ' Set the sign
                                    If striFactor_Current = striFactor_Total Then
                                        strFormula = strFormula & "+"
                                    Else
                                        strFormula = strFormula & "-"
                                    End If
                                    
                                    ' Build the formula
                                    If intRowCount = 0 Then
                                        strFormula = strFormula & "R[" & -(intCurrentRow - intRow_Detail) & "]C"
                                    Else
                                        strFormula = strFormula & "SUM(R[" & -(intCurrentRow - intRow_Detail + intRowCount) & "]C:R[" & -(intCurrentRow - intRow_Detail) & "]C)"
                                    End If
                                    
                                    ' Reset intRowCount 2011-03-16 JPD Added
                                    intRowCount = 0

                                End If
                            
                            End If
                        
                            ' Increment the detail row
                            intRow_Detail = intRow_Detail + 1
                        Next ii
                        
                        ' Assign the formula to all columns in this row with a value of "#Formula-TotalRow#"
                        For ii = intCurrentColumn To intsRowType_Column - 4
                            If .Cells(intCurrentRow, ii).Value = "#Formula-TotalRow#" Then
                                .Cells(intCurrentRow, ii).FormulaR1C1 = strFormula
                            End If
                        Next ii
                        
                    End If ' If intCurrentColumn <> -1
                    
                End If ' If .Cells(intCurrentRow, intsRowType_Column).Value = "T"
                
                ' Increment the current row
                intCurrentRow = intCurrentRow + 1
            Next i

        strLastGoodStatement = "BuildTotalRows(" + strXLS + "," + strReport + ") Completed intCurrentRow = " + CStr(intCurrentRow)

        ' Delete the formula columns
        .Columns(intiFactor_Column).Delete
        .Columns(inthTotalInto_Column).Delete
        .Columns(inthDetail_Column).Delete
        .Columns(intsRowType_Column).Delete

        End With
    
    End If ' If NamedRangeExists(Sheets(strReport), strRange, False)
    
End Sub

Public Sub BuildFormulaR1C1(strXLS As String, strReport As String)
    ' 2009-08-03 JPD Created
    ' 2009-08-12 JPD Added If Not rngFoundCell Is Nothing. Seems to miscount when running interactively.
    ' 2009-08-13 JPD Enhanced to replace |^| with a comma. Embedded commas will be replaced with |^| in the script.
    ' 2010-04-29 JPD Changed xlValues to xlFormulas to include hidden rows and columns
    ' 2013-05-07 JPD Enhanced to replace |^^| with a single quote which is needed to reference other sheets with spaces in their names
    '                Removed replacement of single quote with double quote. Not sure why this was added in 2009.
    '                - We will find out if/when something breaks

    '                   This will create live Excel formulas for the total columns of an account tree
    '                     or any other cell that begins with #FormulaR1C1#
    '                   Requires SQL script that provides properly formatted data

    Dim lngCount As Long
    Dim rngFoundCell As Range

    Set rngFoundCell = Range("A1")
        
    For lngCount = 1 To WorksheetFunction.CountIf(Worksheets(strReport).UsedRange, "#FormulaR1C1#*")

        Set rngFoundCell = Worksheets(strReport).UsedRange.Find(What:="#FormulaR1C1#*", after:=rngFoundCell, _
                LookIn:=xlFormulas, LookAt:=xlPart, SearchOrder:=xlByRows, SearchDirection:=xlNext, MatchCase:=False)

        If Not rngFoundCell Is Nothing Then
            ' Set FormulaR1C1 to the cell value less the leading #FormulaR1C1#. Replace |^| with ,. Launch Excel strips out embeded commas
            rngFoundCell.FormulaR1C1 = Replace(Replace(Replace(rngFoundCell.Value, "#FormulaR1C1#", ""), "|^|", ","), "|^^|", "'")
        Else
            strLastGoodStatement = "BuildFormulaR1C1(" + strXLS + "," + strReport + ")"
            Exit Sub
        End If

    Next lngCount

    strLastGoodStatement = "BuildFormulaR1C1(" + strXLS + "," + strReport + ")"

    Set rngFoundCell = Nothing

End Sub

Public Sub BuildTotalGroups(strXLS As String, strReport As String, strRange As String)
    ' 2009-08-14 JPD Created
    ' 2009-12-08 JPD Changed datatypes from Integer to Long to support larger data sets
    
    '                   This will create live Excel formulas for the total rows of an account tree
    '                   Looks for cells with a value of "#Formula-TotalRow#"
                     
    Dim intCurrentRow As Long
    Dim intRow_Detail As Long
    Dim intCurrentColumn As Long
    Dim i As Long
    Dim ii As Long
    Dim intColumn_Data As Long
    Dim intsRowType_Column As Long
    Dim inthDetail_Column As Long
    Dim inthTotalInto_Column As Long
    Dim intiFactor_Column As Long
    Dim strhTotal As String
    Dim striFactor_Total As String
    Dim striFactor_Current As String
    Dim intRowCount As Long
    
    ' Find the columns and rows we need to identify which rows total into each other
    If NamedRangeExists(Sheets(strReport), strRange, False) Then
        intCurrentRow = Worksheets(strReport).Range(strRange).Row
        intColumn_Data = Worksheets(strReport).Range(strRange).Column
        intsRowType_Column = Worksheets(strReport).UsedRange.Columns.Count - 3
        inthDetail_Column = Worksheets(strReport).UsedRange.Columns.Count - 2
        inthTotalInto_Column = Worksheets(strReport).UsedRange.Columns.Count - 1
        intiFactor_Column = Worksheets(strReport).UsedRange.Columns.Count
        
        With Worksheets(strReport)
            For i = intCurrentRow To .UsedRange.Rows.Count
                strLastGoodStatement = "BuildTotalGroups(" + strXLS + "," + strReport + ") Begin intCurrentRow = " + CStr(intCurrentRow)
                ' Determine if this is a Total row and the row above is not a total row
                If .Cells(intCurrentRow, intsRowType_Column).Value = "T" And _
                    .Cells(intCurrentRow - 1, intsRowType_Column).Value <> "T" Then
                        ' Get the variables for this total row
                        strhTotal = .Cells(intCurrentRow, inthDetail_Column).Value
                        striFactor_Total = .Cells(intCurrentRow, intiFactor_Column).Value
                        ' Set intRow_Detail as the first row of this named range
                        intRow_Detail = Worksheets(strReport).Range(strRange).Row
                        ' Reset variables
                        intRowCount = 0
                    
                        ' Loop through the rows above the current total
                        For ii = intRow_Detail To intCurrentRow - 1
                            ' Get striFactor_Current for this row
                            striFactor_Current = .Cells(intRow_Detail, intiFactor_Column).Value
                            ' Test to see if this detail row totals into intCurrentRow
                            If strhTotal = .Cells(intRow_Detail, inthTotalInto_Column).Value And .Cells(intRow_Detail, intsRowType_Column).Value = "R" Then
                                ' If this is part of a contiguous section then increment the row count
                                If strhTotal = .Cells(intRow_Detail + 1, inthTotalInto_Column).Value And .Cells(intRow_Detail, intsRowType_Column).Value = "R" Then
                                    ' Increment the row count
                                    intRowCount = intRowCount + 1
                                
                                'Otherwise, append this part of the formula
                                Else
                                    Exit For
                                    
                                End If
                            
                            End If
                            
                            ' Increment the detail row
                            intRow_Detail = intRow_Detail + 1
                        Next ii
                        
                        ' Group the rows
                        Worksheets(strReport).Range(intCurrentRow - 1 & ":" & intCurrentRow - intRowCount - 1).Group
                        ' Assign the formula to all columns in this row with a value of "#Formula-TotalRow#"
                        'For ii = intCurrentColumn To intsRowType_Column - 4
                         '   If .Cells(intCurrentRow, ii).Value = "#Formula-TotalRow#" Then
                          '  End If
                        'Next ii
                        
                End If ' If .Cells(intCurrentRow, intsRowType_Column).Value = "T"
                
                ' Increment the current row
                intCurrentRow = intCurrentRow + 1
            Next i

        strLastGoodStatement = "BuildTotalGroups(" + strXLS + "," + strReport + ") Completed intCurrentRow = " + CStr(intCurrentRow)

        End With
    
    End If ' If NamedRangeExists(Sheets(strReport), strRange, False)
    
End Sub

Public Sub BuildReport_Button()
'   2009-08-13 JPD  Created to allow building of report from button on Admin page

    Dim dtStart As Date
    
    If WorksheetExists(ThisWorkbook.Name, "CSV") Then
    
        MsgBox "This process make take a few minutes. Please be patient", vbOKOnly, "Build Report"
        
        dtStart = Now()
        
        With Application
            ' Temporarily disable certain Excel functionality
            .ScreenUpdating = False
            .DisplayAlerts = False
        End With
        
        Call BuildReport(ThisWorkbook.Name)
    
        With Application
            ' Enable certain Excel functionality
            .ScreenUpdating = True
            .DisplayAlerts = True
        End With
        
        MsgBox "Process complete." & vbCrLf & "Start: " & dtStart & vbCrLf & "End: " & Now(), vbOKOnly, "Build Report"
    
    Else
        MsgBox "CSV worksheet not found. Report cannot be built.", vbOKOnly, "Build Report"

    End If

End Sub
Public Sub DynamicVertical(strXLS As String, strReport As String, strRangeName As String, intRows As Long)
    ' 2009-09-29 JPD Created
    ' 2009-12-22 JPD Added support for intColumns_Template <> 0 and copying formatting
    
    '                   Range names that begin with v_ have the ability to grow dynamically
    '                   Range name is in the following format: v_RXX_CYY_MyNamedRange
    '                    - XX number of blank rows in the template. If iRows > XX then we need to insert rows
    '                    - YY number of columns in the range. 00 means we will insert an entire row
    Dim intRows_Template As Long
    Dim intColumns_Template As Long
    Dim i As Long
 
    strLastGoodStatement = "Start DynamicVertical (Report :" & strReport & ", Range: " & strRangeName & ")"
    
    intRows_Template = Mid(strRangeName, 4, 2)
    intColumns_Template = Mid(strRangeName, 8, 2)

    If intRows > intRows_Template Then
        For i = 1 To intRows - intRows_Template
            If intColumns_Template = 0 Then
                ' Insert an entire row
                With Worksheets(strReport)
                    .Cells(.Range(strRangeName).Row + i, _
                        .Range(strRangeName).Column).EntireRow.Insert
                ' Format the inserted row the same as the source row
                .Cells(.Range(strRangeName).Row, _
                        .Range(strRangeName).Column).EntireRow.Copy
                .Cells(.Range(strRangeName).Row + i, _
                        .Range(strRangeName).Column).EntireRow.PasteSpecial Paste:=xlPasteFormats
                End With
            Else
                ' Insert specific number of columns
                With Worksheets(strReport)
                    .Range( _
                        .Cells(.Range(strRangeName).Row + i, _
                            .Range(strRangeName).Column), _
                        .Cells(.Range(strRangeName).Row + i, _
                            .Range(strRangeName).Column + intColumns_Template - 1) _
                            ).Insert Shift:=xlDown
                ' Format the inserted row the same as the source row
                    .Range( _
                        .Cells(.Range(strRangeName).Row, _
                            .Range(strRangeName).Column), _
                        .Cells(.Range(strRangeName).Row, _
                            .Range(strRangeName).Column + intColumns_Template - 1) _
                            ).Copy
                    .Range( _
                        .Cells(.Range(strRangeName).Row + i, _
                            .Range(strRangeName).Column), _
                        .Cells(.Range(strRangeName).Row + i, _
                            .Range(strRangeName).Column + intColumns_Template - 1) _
                            ).PasteSpecial Paste:=xlPasteFormats
                    
                End With
            End If
        Next i
    Else
        ' There are already enough rows in the template, so there is nothing to do.
    End If

    strLastGoodStatement = "Complete DynamicVertical (Report :" & strReport & ", Range: " & strRangeName & ")"

End Sub

Public Sub FormatCells(strXLS As String, strReport As String, strPrefix As String, strFormatSheet As String)
    ' 2010-04-28 JPD Created to support formatting cells based on predefined rules or named ranges in the Format tab
    ' 2010-08-25 JPD Added support for RowDelete and ColumnDelete
    ' 2011-04-14 JPD Added logic to check that xlInsideHorizontal and xlInsideVertical have 2 rows
    ' 2011-05-13 JPD Added support for RowGroup, RowUngroup, ColumnGroup, ColumnUngroup
    ' 2011-09-23 JPD Added support for HPageBreak, VPageBreak and Value
    ' 2011-10-07 JPD Added rngDestination.Calculate to Value
    ' 2012-02-17 JPD Added Excel 2000 compatibility for SheetProtect
    ' 2012-07-30 JPD Added support for attributes for Underline
    ' 2012-08-28 JPD Added support for ChartMinimum. (2013-03-02 This was replaced by ChartValueAxisMinimum)
    ' 2013-03-02 JPD Added support for ChartValueAxisMinimum and ChartValueAxisMaximum
    
    ' strPrefix = #FC_, #FCAfter_ or #FCFinal_ depending on when it is called
    
    ' This procedure formats cells based on keywords in this format #FC/FCAfter_Predefined/NamedRange_AttributeAttribute_RX_CX#
    ' 1. Must start with #FC_ or #FCAfter_ (this syntax is reserved for the FormatCells function)
    ' 2. Predefined format or named range on the Format tab that is the source of the attributes we will copy
    '    -  Must not have an underscore within the named range's name
    ' 3. Listing of one or more of Attributes which are the tab names from the Format Cells window
    '    - Number/Alignment/Font/Border/Fill/Protection
    '       - Can abbreviate with first three of more characters of each
    '       - Can separate with comma, slash, or nothing at all
    '         - Num,Ali,Fon,Bor,Fil,Pro OR Num/Ali/Fon/Bor/Fil/Pro OR NumAliFonBorFilPro
    '    - All (This will result in a Paste Special-->Formats)
    '    Attributes must appear between the underscore characters
    '    For predefined formats the attribute represents the color, size or other formatting option
    ' 4. OPTIONAL - _RXX
    '    - XX = Number of rows that are included as the destination for the format
    '    - If this parameter is omitted the default is 01
    ' 5. OPTIONAL - _CXX
    '    - XX = Number of columns that are included as the destination for the format
    '    - If this parameter is omitted the default is 01
    '    - 00 = all columns from the current cell to the end of the worksheet's used range
    ' 6. Must end with # and have not other # characters in this keyword
    ' 7. FormatCells keywords can located anywhere in the cell's text
    ' 8. Multiple FormatCells keywords in the same cell will be processed from left to right
      
    Dim rng As Range
    Dim boolFound As Boolean
    Dim iStart As Long
    Dim iEnd As Long
    Dim iEnd_Prior As Long
    Dim strKeyword As String
    Dim strAttributes As String
    Dim strRangeName As String
    Dim intRows As Long
    Dim intColumns As Long
    Dim rngSource As Range
    Dim rngDestination As Range
    Dim strErrorMessage As String
    Dim BordersIndex As XlBordersIndex
    Dim i As Long
    Dim strColor As String
    Dim LineStyle_Attribute As XlLineStyle
    Dim LineStyle_Current As XlLineStyle
    Dim BorderWeight_Attribute As XlBorderWeight
    Dim BorderWeight_Current As XlBorderWeight
    Dim Row As Range
    
    ' Initialize variable
    boolFound = True
    
    Do While boolFound = True
        With Workbooks(strXLS).Worksheets(strReport).UsedRange
            Set rng = .Find(What:="*" + strPrefix + "*#*", _
                            after:=.Cells(.Cells.Count), _
                            LookIn:=xlFormulas, _
                            LookAt:=xlWhole, _
                            SearchOrder:=xlByRows, _
                            SearchDirection:=xlNext, _
                            MatchCase:=False)
            If Not rng Is Nothing Then
                ' Get the start of the keyword
                iStart = InStr(1, rng.Value, strPrefix)
                ' Get the end of the keyword
                iEnd = InStr(iStart + Len(strPrefix), rng.Value, "#")
                ' Parse it to find the keyword.
                strKeyword = Mid(rng.Value, iStart, iEnd - iStart + 1)
                ' Set the start of the Range Name which come after the Prefix
                iStart = Len(strPrefix) + 1
                ' Get the end of the Range Name
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if RangeName is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the Range Name
                If iEnd > iStart Then
                    strRangeName = Mid(strKeyword, iStart, iEnd - iStart)
                Else
                    strRangeName = ""
                End If
                ' Set the start of the Attributes which come after the Range Name
                iStart = iEnd + 1
                ' Get the end of the Attributes
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if Attributes are present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the Attributes
                If iEnd > iStart Then
                    strAttributes = Mid(strKeyword, iStart, iEnd - iStart)
                Else
                    strAttributes = ""
                End If
                ' Set the start of the Rows which may come after the Attributes
                iStart = InStr(iEnd_Prior, strKeyword, "_R")
                If iStart > 0 Then
                    ' Set the start of the numeric portion
                    iStart = iStart + 2
                    ' Get the end of the numeric portion
                    ' First look for the underscore
                    iEnd = InStr(iStart, strKeyword, "_")
                    If iEnd = 0 Then
                        ' Otherwise look for the hash
                        iEnd = InStr(iStart, strKeyword, "#")
                    End If
                    If iEnd > iStart Then
                        If IsNumeric(Mid(strKeyword, iStart, iEnd - iStart)) Then
                          intRows = Mid(strKeyword, iStart, iEnd - iStart)
                        Else
                          intRows = 1
                        End If
                    Else
                        'Default to 1
                        intRows = 1
                    End If
                Else
                    'Default to 1
                    intRows = 1
                End If
                ' Set the start of the Columns which may come after the Range Name
                iStart = InStr(iEnd_Prior, strKeyword, "_C")
                If iStart > 0 Then
                    ' Set the start of the numeric portion
                    iStart = iStart + 2
                    ' Get the end of the numeric portion
                    ' First look for the underscore
                    iEnd = InStr(iStart, strKeyword, "_")
                    If iEnd = 0 Then
                        ' Otherwise look for the hash
                        iEnd = InStr(iStart, strKeyword, "#")
                    End If
                    If iEnd > iStart Then
                        If IsNumeric(Mid(strKeyword, iStart, iEnd - iStart)) Then
                          intColumns = Mid(strKeyword, iStart, iEnd - iStart)
                        Else
                          intColumns = 1
                        End If
                    Else
                        'Default to 1
                        intColumns = 1
                    End If
                Else
                    'Default to 1
                    intColumns = 1
                End If
                
                ' Sheet is protected
                If rng.Parent.ProtectContents Then
                    ' If the first attempt isn't unprotecting the sheet then it's time to leave
                    If strRangeName <> "SheetUnprotect" Then
                        Call LogError("FormatCells", "Sheet protected. Could not process (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                        GoTo ExitProcedure
                    End If
                Else
                    ' Remove the keyword from the cell value
                    rng.Value = Replace(rng.Value, strKeyword, "")
                End If
                        
                ' Confirm the rows
                If intRows = 0 Then
                    ' Extend from the current row to the end of the used range
                    intRows = rng.Parent.UsedRange.Rows.Count - rng.Row + 1
                End If
                ' Confirm the columns
                If intColumns = 0 Then
                    ' Extend from the current column to the end of the used range
                    intColumns = rng.Parent.UsedRange.Columns.Count - rng.Column + 1
                End If
                
                ' Set the destination range
                Set rngDestination = rng.Parent.Range(rng.Parent.Cells(rng.Row, rng.Column), _
                    rng.Parent.Cells(rng.Row + intRows - 1, rng.Column + intColumns - 1))
                
    ' **** Begin "Virtual" RangeNames ****
                ' Test to see if strRangeName is any of a number of reserved "virtual" RangeNames
                ' Formatting is "hard coded" so the Format worksheet is not required
                If strRangeName = "NoBorder" Then
                    rngDestination.Borders.LineStyle = xlNone
                ElseIf strRangeName = "Border" Then
                    ' Supported strAttributes are: All, Outline, Top, Bottom, Left, Right, DiagonalUp, DiagonalDown, 1, 2, 3
                    ' 1 = Single Overline, 2 = Single Overline and Double Underline, 3 = Single Underline
                    ' Default LineStyle is Continuous. Additional options are Dash, DashDot, DashDotDot, Dot, Double, SlantDashDot
                    If strAttributes Like "*SlantDashDot*" Then
                        LineStyle_Attribute = xlSlantDashDot
                    ElseIf strAttributes Like "*DashDotDot*" Then
                        LineStyle_Attribute = xlDashDotDot
                    ElseIf strAttributes Like "*DashDot*" Then
                        LineStyle_Attribute = xlDashDot
                    ElseIf strAttributes Like "*Dash*" Then
                        LineStyle_Attribute = xlDash
                    ElseIf strAttributes Like "*Dot*" Then
                        LineStyle_Attribute = xlDot
                    ElseIf strAttributes Like "*Double*" Then
                        LineStyle_Attribute = xlDouble
                    Else
                        LineStyle_Attribute = xlContinuous
                    End If
                    
                    ' Default BorderWeight is Thin. Additional options are Medium, Thick, Hairline.
                    If strAttributes Like "*Medium*" Then
                        BorderWeight_Attribute = xlMedium
                    ElseIf strAttributes Like "*Thick*" Then
                        BorderWeight_Attribute = xlThick
                    ElseIf strAttributes Like "*Hairline*" Then
                        BorderWeight_Attribute = xlHairline
                    Else
                        BorderWeight_Attribute = xlThin
                    End If
    
                   ' Loop through XLBordersIndex values
                    For i = 1 To 8
                        ' Reset current value to attribute
                        LineStyle_Current = LineStyle_Attribute
                        BorderWeight_Current = BorderWeight_Attribute
                        
                        Select Case i
                            Case 1
                                BordersIndex = xlDiagonalDown
                                ' Override for exceptions
                                If Not strAttributes Like "*DiagonalDown*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                            Case 2  ' Diagonal is not supported
                                BordersIndex = xlDiagonalUp
                                ' Override for exceptions
                                If Not strAttributes Like "*DiagonalUp*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                            Case 3
                                BordersIndex = xlEdgeBottom
                                ' Override for exceptions
                                If strAttributes Like "*2*" Then ' Double underline
                                    LineStyle_Current = xlDouble
                                ElseIf Not strAttributes Like "*All*" And _
                                    Not strAttributes Like "*Outline*" And _
                                    Not strAttributes Like "*Bottom*" And _
                                    Not strAttributes Like "*3*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                            Case 4
                                BordersIndex = xlEdgeLeft
                                ' Override for exceptions
                                If Not strAttributes Like "*All*" And _
                                    Not strAttributes Like "*Outline*" And _
                                    Not strAttributes Like "*Left*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                            Case 5
                                BordersIndex = xlEdgeRight
                                ' Override for exceptions
                                If Not strAttributes Like "*All*" And _
                                    Not strAttributes Like "*Outline*" And _
                                    Not strAttributes Like "*Right*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                            Case 6
                                BordersIndex = xlEdgeTop
                                ' Override for exceptions
                                If Not strAttributes Like "*All*" And _
                                    Not strAttributes Like "*Outline*" And _
                                    Not strAttributes Like "*Top*" And _
                                    Not strAttributes Like "*1*" And _
                                    Not strAttributes Like "*2*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                            Case 7
                                BordersIndex = xlInsideHorizontal
                                ' Override for exceptions
                                If Not strAttributes Like "*All*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                            Case 8
                                BordersIndex = xlInsideVertical
                                ' Override for exceptions
                                If Not strAttributes Like "*All*" Then
                                    LineStyle_Current = xlLineStyleNone
                                End If
                        End Select
                        ' Apply the formatting
                        With rngDestination.Borders(BordersIndex)
                            ' 2010-06-01 JPD Excel 2003 requires 2 rows for xlInsideHorizontal and 2 columns for xlInsideVertical
                            If i = 7 And rngDestination.Rows.Count < 2 Then
                            ElseIf i = 8 And rngDestination.Columns.Count < 2 Then
                            Else
                                .ColorIndex = xlAutomatic
                                .Weight = BorderWeight_Current    ' Weight needs to come before LineStyle
                                .LineStyle = LineStyle_Current
                            End If
                        End With
                    Next i
                ' *** End strRangeName = "Border"
                ElseIf strRangeName = "Bold" Then
                    rngDestination.Font.Bold = True
                ElseIf strRangeName = "NoBold" Then
                    rngDestination.Font.Bold = False
                ElseIf strRangeName = "Underline" Then
                    ' 2012-07-30 JPD Added support for attributes for Underline
                    If strAttributes Like "*S*A*" Then
                        rngDestination.Font.Underline = xlUnderlineStyleSingleAccounting
                    ElseIf strAttributes Like "*D*A*" Then
                        rngDestination.Font.Underline = xlUnderlineStyleDoubleAccounting
                    ElseIf strAttributes Like "*S*" Then
                        rngDestination.Font.Underline = xlUnderlineStyleSingle
                    ElseIf strAttributes Like "*D*" Then
                        rngDestination.Font.Underline = xlUnderlineStyleDouble
                    Else
                        rngDestination.Font.Underline = True
                    End If
                ElseIf strRangeName = "NoUnderline" Then
                    rngDestination.Font.Underline = False
                ElseIf strRangeName = "Italic" Then
                    rngDestination.Font.Italic = True
                ElseIf strRangeName = "NoItalic" Then
                    rngDestination.Font.Italic = False
                ElseIf strRangeName = "Lock" Then
                    rngDestination.Locked = True
                ElseIf strRangeName = "NoLock" Then
                    rngDestination.Locked = False
                ElseIf strRangeName = "Value" Then
                    rngDestination.Calculate ' 2011-10-07 JPD Added
                    rngDestination.Value = rngDestination.Value
                ElseIf strRangeName = "FontColor" Then
                    ' Find the color from the attribute. Must be 9 digits in RGB format: RRRGGGBBB
                    If IsNumeric(strAttributes) And Len(strAttributes) = 9 Then
                        rngDestination.Font.Color = RGB(Left(strAttributes, 3), Mid(strAttributes, 4, 3), Right(strAttributes, 3))
                    Else
                        Call LogError("FormatCells", "Invalid RGB color (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ElseIf strRangeName = "FontSize" Then
                    ' Find the size from the attribute. Must be numeric.
                    If IsNumeric(strAttributes) Then
                        ' Must be between 0 and 409
                        If strAttributes >= 0 And strAttributes <= 409 Then
                                rngDestination.Font.Size = strAttributes
                        Else
                            Call LogError("FormatCells", "Invalid Font Size (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                        End If
                    Else
                        Call LogError("FormatCells", "Invalid Font Size (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ElseIf strRangeName = "FillColor" Then
                    ' Find the color from the attribute. Must be 9 digits in RGB format: RRRGGGBBB
                    If IsNumeric(strAttributes) And Len(strAttributes) = 9 Then
                        rngDestination.Interior.Color = RGB(Left(strAttributes, 3), Mid(strAttributes, 4, 3), Right(strAttributes, 3))
                    Else
                        Call LogError("FormatCells", "Invalid RGB color (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ElseIf strRangeName = "RowHide" Then
                    rngDestination.Rows.Hidden = True
                ElseIf strRangeName = "RowUnhide" Then
                    rngDestination.Rows.Hidden = False
                ElseIf strRangeName = "RowGroup" Then
                    rngDestination.Rows.Group
                ElseIf strRangeName = "RowUngroup" Then
                    If IsNull(rngDestination.Rows.OutlineLevel) = 1 Or rngDestination.Rows.OutlineLevel <> 1 Then
                        rngDestination.Rows.Ungroup
                    End If
                ElseIf strRangeName = "RowDelete" Then
                    rngDestination.Rows.EntireRow.Delete
                ElseIf strRangeName = "RowAutoFit" Then
                    rngDestination.Rows.AutoFit
                    Call RowsAutoFit(rngDestination)
                ElseIf strRangeName = "RowHeight" Then
                    ' Find the height from the attribute. Must be numeric.
                    If IsNumeric(strAttributes) Then
                        ' Must be between 0 and 409
                        If strAttributes >= 0 And strAttributes <= 409 Then
                                rngDestination.Rows.RowHeight = strAttributes
                        Else
                            Call LogError("FormatCells", "Invalid Row Height (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                        End If
                    Else
                        Call LogError("FormatCells", "Invalid Row Height (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ElseIf strRangeName = "HPageBreak" Then
                    If strAttributes Like "A*" Then
                        rngDestination.Parent.HPageBreaks.Add (Cells(rngDestination.Row + 1, rngDestination.Column))
                    Else
                        ' Default Attribute is Before
                        rngDestination.Parent.HPageBreaks.Add (Cells(rngDestination.Row, rngDestination.Column))
                    End If
                ElseIf strRangeName = "VPageBreak" Then
                    If strAttributes Like "A*" Then
                        rngDestination.Parent.VPageBreaks.Add (Cells(rngDestination.Row, rngDestination.Column + 1))
                    Else
                        ' Default Attribute is Before
                        rngDestination.Parent.VPageBreaks.Add (Cells(rngDestination.Row, rngDestination.Column))
                    End If
                ElseIf strRangeName = "ColumnHide" Then
                    rngDestination.Columns.Hidden = True
                ElseIf strRangeName = "ColumnUnhide" Then
                    rngDestination.Columns.Hidden = False
                ElseIf strRangeName = "ColumnGroup" Then
                    rngDestination.Columns.Group
                ElseIf strRangeName = "ColumnUngroup" Then
                    If IsNull(rngDestination.Columns.OutlineLevel) = 1 Or rngDestination.Columns.OutlineLevel <> 1 Then
                        rngDestination.Columns.Ungroup
                    End If
                ElseIf strRangeName = "ColumnDelete" Then
                    rngDestination.Columns.EntireColumn.Delete
                ElseIf strRangeName = "ColumnAutoFit" Then
                    rngDestination.Columns.EntireColumn.AutoFit
                ElseIf strRangeName = "ColumnWidth" Then
                    ' Find the width from the attribute. Must be numeric.
                    If IsNumeric(strAttributes) Then
                        ' Must be between 0 and 255
                        If strAttributes >= 0 And strAttributes <= 255 Then
                            rngDestination.Columns.ColumnWidth = strAttributes
                        Else
                            Call LogError("FormatCells", "Invalid Column Width (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                        End If
                    Else
                        Call LogError("FormatCells", "Invalid Column Width (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ElseIf strRangeName = "SheetTabColor" Then
                    ' Find the color from the attribute. Must be 9 digits in RGB format: RRRGGGBBB
                    If IsNumeric(strAttributes) And Len(strAttributes) = 9 Then
                        rngDestination.Parent.Tab.Color = RGB(Left(strAttributes, 3), Mid(strAttributes, 4, 3), Right(strAttributes, 3))
                    Else
                        Call LogError("FormatCells", "Invalid RGB color (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ElseIf strRangeName Like "ChartValueAxisMinimum*" Then
                    ' Find the chart name from the attribute.
                    If IsNumeric(strAttributes) Then
                        rngDestination.Parent.ChartObjects(Replace(strRangeName, "ChartValueAxisMinimum", "")).Chart.Axes(xlValue).MinimumScale = strAttributes
                    Else
                        Call LogError("FormatCells", "Invalid ChartValueAxisMinimum value (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ElseIf strRangeName Like "ChartValueAxisMaximum*" Then
                    ' Find the chart name from the attribute.
                    If IsNumeric(strAttributes) Then
                        rngDestination.Parent.ChartObjects(Replace(strRangeName, "ChartValueAxisMaximum", "")).Chart.Axes(xlValue).MaximumScale = strAttributes
                    Else
                        Call LogError("FormatCells", "Invalid ChartValueAxisMaximum value (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                    End If
                ' This better be the last action taken on the sheet
                ElseIf strRangeName = "SheetProtect" Then
                    ' 2012-02-17 JPD Added Excel 2000 compatibility
                    If Application.Version <= 9 Then    ' Excel 2000
                        rngDestination.Parent.Protect Password:=strAttributes, DrawingObjects:=True, Contents:=True, Scenarios:=True
                    Else
                        rngDestination.Parent.Protect Password:=strAttributes, DrawingObjects:=True, Contents:=True, Scenarios:=True, _
                            AllowFormattingCells:=True, AllowFormattingColumns:=True, AllowFormattingRows:=True, _
                            AllowInsertingColumns:=True, AllowInsertingRows:=True, AllowDeletingColumns:=True, _
                            AllowDeletingRows:=True
                    End If
                    
                    ' After the sheet has been protected it's time to leave
                    GoTo ExitProcedure
                ' This better be the first action taken on the sheet
                ElseIf strRangeName = "SheetUnprotect" Then
                    rngDestination.Parent.Unprotect Password:=strAttributes
                    ' Remove the keyword from the cell value
                    rng.Value = Replace(rng.Value, strKeyword, "")
    ' **** End "Virtual" RangeNames ****
                ' Test to see if a formatting worksheet exists
                ElseIf WorksheetExists(strXLS, strFormatSheet, True) Then
                    ' Test to see that the named range exists
                    If NamedRangeExists(Workbooks(strXLS).Worksheets(strFormatSheet), strRangeName, True, False) Then
                        ' Set the source range
                        Set rngSource = Workbooks(strXLS).Worksheets(strFormatSheet).Range(strRangeName)
                        
                        ' Step through each attribute and apply the formatting
                        With rngDestination
                            If strAttributes Like "*Num*" Then 'Number
                                .NumberFormat = rngSource.NumberFormat
                            End If
                            
                            If strAttributes Like "*Ali*" Then 'Alignment
                                .HorizontalAlignment = rngSource.HorizontalAlignment
                                .IndentLevel = rngSource.IndentLevel
                                .VerticalAlignment = rngSource.VerticalAlignment
                                .WrapText = rngSource.WrapText
                                .ShrinkToFit = rngSource.ShrinkToFit
                                If rngSource.MergeCells = True Then
                                    ' If we are going to merge, be certain rngDestination is not already merged
                                    .MergeCells = False
                                End If
                                .MergeCells = rngSource.MergeCells
                                .Orientation = rngSource.Orientation
                            End If
                        
                            If strAttributes Like "*Fon*" Then 'Font
                                With .Font
                                    .Bold = rngSource.Font.Bold
                                    .Color = rngSource.Font.Color
                                    '.ColorIndex = rngSource.Font.ColorIndex
                                    .FontStyle = rngSource.Font.FontStyle
                                    .Italic = rngSource.Font.Italic
                                    .Name = rngSource.Font.Name
                                    .Size = rngSource.Font.Size
                                    .Strikethrough = rngSource.Font.Strikethrough
                                    .Subscript = rngSource.Font.Subscript
                                    .Superscript = rngSource.Font.Superscript
                                    .Underline = rngSource.Font.Underline
                                    ' Properties not supported prior to Excel 2007
                                    If Application.Version >= 12 Then ' Excel 2007
                                        .TintAndShade = rngSource.Font.TintAndShade
                                        '.ThemeColor = rngSource.Font.ThemeColor
                                        .ThemeFont = rngSource.Font.ThemeFont
                                    End If
                                End With
                            End If
                            
                            If strAttributes Like "*Bor*" Then 'Border
                                ' Loop through XLBordersIndex values
                                For i = 1 To 8
                                    Select Case i
                                        Case 1
                                            BordersIndex = xlDiagonalDown
                                        Case 2
                                            BordersIndex = xlDiagonalUp
                                        Case 3
                                            BordersIndex = xlEdgeBottom
                                        Case 4
                                            BordersIndex = xlEdgeLeft
                                        Case 5
                                            BordersIndex = xlEdgeRight
                                        Case 6
                                            BordersIndex = xlEdgeTop
                                        Case 7
                                            BordersIndex = xlInsideHorizontal
                                        Case 8
                                            BordersIndex = xlInsideVertical
                                    End Select
                                    ' Apply the formatting
                                    With .Borders(BordersIndex)
                                        ' 2011-04-14 JPD Excel 2003 requires 2 rows for xlInsideHorizontal and 2 columns for xlInsideVertical
                                        If i = 7 And rngDestination.Rows.Count < 2 Then
                                        ElseIf i = 8 And rngDestination.Columns.Count < 2 Then
                                        Else
                                            .Color = rngSource.Borders(BordersIndex).Color
                                            '.ColorIndex = rngSource.Borders(BordersIndex).ColorIndex
                                            .Weight = rngSource.Borders(BordersIndex).Weight    ' Weight needs to come before LineStyle
                                            .LineStyle = rngSource.Borders(BordersIndex).LineStyle
                                        End If
                                        
                                        ' Properties not supported prior to Excel 2007
                                        If Application.Version >= 12 Then ' Excel 2007
                                            .TintAndShade = rngSource.Borders(BordersIndex).TintAndShade
                                        End If
                                        
                                    End With
                                Next i
                            End If
                            
                            If strAttributes Like "*Fil*" Then 'Fill
                                With .Interior
                                    .Color = rngSource.Interior.Color
                                    '.ColorIndex = rngSource.Interior.ColorIndex
                                    '.InvertIfNegative = rngSource.Interior.InvertIfNegative
                                    .Pattern = rngSource.Interior.Pattern
                                    .PatternColor = rngSource.Interior.PatternColor
                                    .PatternColorIndex = rngSource.Interior.PatternColorIndex
                                    ' Properties not supported prior to Excel 2007
                                    If Application.Version >= 12 Then ' Excel 2007
                                        '.Gradient = rngSource.Interior.Gradient
                                        '.PatternThemeColor = rngSource.Interior.PatternThemeColor
                                        .PatternTintAndShade = rngSource.Interior.PatternTintAndShade
                                        '.ThemeColor = rngSource.Interior.ThemeColor
                                        .TintAndShade = rngSource.Interior.TintAndShade
                                    End If
                                End With
                            End If
                            
                            If strAttributes Like "*Pro*" Then 'Protection
                                .Locked = rngSource.Locked
                            End If
                            
                            If strAttributes Like "*All*" Then 'All
                                If rngSource.MergeCells = True Then
                                    ' If we are going to merge, be certain rngDestination is not already merged
                                    .MergeCells = False
                                End If
                                rngSource.Copy
                                rngDestination.PasteSpecial xlPasteFormats
                            End If
                        
                        End With
                        
                        strLastGoodStatement = "Worksheets(" + strReport + ") Format Cells(" + rng.Address + ")"
                    
                    End If
                
                End If
                
                ' If we have made one pass and the sheet is still locked then exit. There is nothing more we can do.
                If Workbooks(strXLS).Worksheets(strReport).ProtectContents Then
                    ' Stop the loop
                    boolFound = False
                    Call LogError("FormatCells", "Sheet protected. Could not process (" & strAttributes & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                End If

            Else
                ' Stop the loop
                boolFound = False
            End If
        End With
    Loop
    
ExitProcedure:
    Set rng = Nothing
    Set rngSource = Nothing
    Set rngDestination = Nothing
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("FormatCells", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "FormatCells"
            #End If
            
    End Select
    Resume ExitProcedure

End Sub

Public Sub TextFile(strXLS As String, strReport As String, strPrefix As String, strFormatSheet As String)
    ' 2010-12-01 JPD Created to support embedding a text file in the workbook
    ' 2011-02-03 JPD Correct .Activate code prior to Set myEmbededShape to properly embed files on different Worksheet than source data
    ' strPrefix = #TextFile_
    ' strFormatSheet has not been implemented yet
    
    ' This procedure embeds a text file in the cell with keywords in this format:
    '        #TextFile_Delimiter_FileName_SourceReport_SourceRangeName_RX_CX_DeleteSourceReport_ValuesOrFormulas_MyFormatShape#
    ' 1. Must start with #TextFile_ (this syntax is reserved for the TextFile function)
    ' 2. Delimiter
    '    - Comma/Tab
    '      - Can abbreviate with first character (C or T)
    '    - Comma = Adds quotes around cell values and separate with a comma
    '    - Tab = Separate cell values with a tab
    ' 3. OPTIONAL - FileName
    '    - Name for the embedded text file
    '    - Default is MyFile.txt
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores in the worksheet name
    '    - Use placeholder underscore if this parameter is omitted
    ' 4. SourceReport
    '    - Name of worksheet with data
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores in the worksheet name
    ' 5. SourceRangeName
    '    - Named range on SourceReport that has the data
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores in the named range
    ' 6. OPTIONAL - _RX
    '    - X = Number of rows to include, starting with the first cell in the SourceRangeName
    '    - 0 = All rows from start of SourceRangeName to end of SourceReport used range
    '    - If this parameter is omitted the default is 0
    '    - Use placeholder underscore if this parameter is omitted and other subsequent parameters are used
    ' 7. OPTIONAL - _CX
    '    - X = Number of columns to include, starting with the first cell in the SourceRangeName
    '    - 0 = All columns from start of SourceRangeName to end of SourceReport used range
    '    - If this parameter is omitted the default is 0
    '    - Use placeholder underscore if this parameter is omitted and other subsequent parameters are used
    ' 8. OPTIONAL - _DeleteSourceReport
    '    - YesDelete/NoDelete
    '      - Can abbreviate with first character (Y or N)
    '    - If this parameter is omitted the default is N
    '    - If YesDelete, SourceReport worksheet is deleted after file is created
    '    - Use placeholder underscore if this parameter is omitted and other subsequent parameters are used
    ' 9. OPTIONAL - _ValuesOrFormulas
    '    - Values/Formulas
    '      - Can abbreviate with first character (V or F)
    '    - If this parameter is omitted the default is V
    '    - Values = the text file will have the values from each cell
    '    - Formulas = the text file will have the formulas from each cell
    '    - Use placeholder underscore if this parameter is omitted and other subsequent parameters are used
    '10. OPTIONAL - _MyFormatShape
    '    - Name of a shape object on the Format tab to copy formatting from
    '    - If this parameter is omitted the default is MyFormatShape
    '    - If the shape object cannot be found in the Format worksheet, detail formats will be applied
    '      - .Fill.ForeColor.RGB = 65535 ' Yellow
    '      - .Height = 150
    '      - .Width = 300
    '      - .Line.ForeColor.RGB = 0
    '      - .Line.BackColor.RGB = 0
    '      - .Line.Weight = 1
    '      - .Line.Style = 1
    
    
    Dim rng As Range
    Dim boolFound As Boolean
    Dim iStart As Long
    Dim iEnd As Long
    Dim iEnd_Prior As Long
    Dim strKeyword As String
    Dim strDelimiter As String
    Dim strFileName As String
    Dim strSourceReport As String
    Dim strSourceRangeName As String
    Dim intRows As Long
    Dim intColumns As Long
    Dim strDeleteSourceReport As String
    Dim strValuesOrFormulas As String
    Dim strMyFormatShape As String
    Dim rngSource As Range
    Dim strErrorMessage As String
    Dim strOutput As String
    Dim strOutputFile As String
    Dim iFile As Long
    Dim strDelimiterCharacter As String
    Dim Row As Range
    Dim Cell As Range
    Dim myFormatShape As Shape
    Dim myEmbededShape As Shape
        
    On Error GoTo ErrorHandler
        
    ' Initialize variable
    boolFound = True
    
    Do While boolFound = True
        With Workbooks(strXLS).Worksheets(strReport).UsedRange
            Set rng = .Find(What:="*" + strPrefix + "*#*", _
                            after:=.Cells(.Cells.Count), _
                            LookIn:=xlFormulas, _
                            LookAt:=xlWhole, _
                            SearchOrder:=xlByRows, _
                            SearchDirection:=xlNext, _
                            MatchCase:=False)
            If Not rng Is Nothing Then
               strLastGoodStatement = "TextFile parsing begin Worksheets(" + strReport + ") Range(" + rng.Address + ")"
                ' Get the start of the keyword
                iStart = InStr(1, rng.Value, strPrefix)
                ' Get the end of the keyword
                iEnd = InStr(iStart + Len(strPrefix), rng.Value, "#")
                ' Parse it to find the keyword.
                strKeyword = Mid(rng.Value, iStart, iEnd - iStart + 1)
                ' Set the start of the Delimiter which come after the Prefix
                iStart = Len(strPrefix) + 1
                ' Get the end of the Delimiter
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if Delimiter is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the Delimiter
                If iEnd > iStart Then
                    strDelimiter = Mid(strKeyword, iStart, iEnd - iStart)
                Else
                    strDelimiter = ""
                End If
                
                iStart = iEnd_Prior + 1
                ' If the next character is an underscore, this optional parameter was skipped
                If Mid(strKeyword, iEnd_Prior, 2) = "__" Then
                    iEnd = iStart
                    iEnd_Prior = iEnd
                    strFileName = "MyFile.txt"
                Else
                    ' Get the end of the FileName
                    ' First look for the underscore
                    iEnd = InStr(iStart, strKeyword, "_")
                    If iEnd = 0 Then
                        ' Otherwise look for the hash
                        iEnd = InStr(iStart, strKeyword, "#")
                    End If
                    ' Update iEnd_Prior if FileName is present
                    If iEnd > 0 Then iEnd_Prior = iEnd
                    ' Parse it to find the FileName
                    If iEnd > iStart Then
                        strFileName = Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_")
                    Else
                        strFileName = "MyFile.txt"
                    End If
                End If
                
                ' Set the start of the SourceReport which comes after the FileName
                iStart = iEnd_Prior + 1
                ' Get the end of the SourceReport
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if SourceReport is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the SourceReport
                If iEnd > iStart Then
                    strSourceReport = Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_")
                Else
                    strSourceReport = ""
                End If
                
                ' Set the start of the SourceRangeName which comes after the SourceReport
                iStart = iEnd + 1
                ' Get the end of the SourceRangeName
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if SourceRangeName is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the SourceRangeName
                If iEnd > iStart Then
                    strSourceRangeName = Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_")
                Else
                    strSourceRangeName = ""
                End If
                
                iStart = iEnd_Prior + 1
                ' If the next character is an underscore, this optional parameter was skipped
                If Mid(strKeyword, iEnd_Prior, 2) = "__" Then
                    iEnd = iStart
                    iEnd_Prior = iEnd
                    intRows = 0
                Else
                    ' Set the start of the Rows which may come after the SourceRangeName
                    iStart = InStr(iEnd_Prior, strKeyword, "_R")
                    If iStart > 0 Then
                        ' Set the start of the numeric portion
                        iStart = iStart + 2
                        ' Get the end of the numeric portion
                        ' First look for the underscore
                        iEnd = InStr(iStart, strKeyword, "_")
                        If iEnd = 0 Then
                            ' Otherwise look for the hash
                            iEnd = InStr(iStart, strKeyword, "#")
                        End If
                        ' Update iEnd_Prior if Columns is present
                        If iEnd > 0 Then iEnd_Prior = iEnd
                        If iEnd > iStart Then
                            If IsNumeric(Mid(strKeyword, iStart, iEnd - iStart)) Then
                              intRows = Mid(strKeyword, iStart, iEnd - iStart)
                            Else
                              intRows = 0
                            End If
                        Else
                            'Default to 0
                            intRows = 0
                        End If
                    Else
                        'Default to 0
                        intRows = 0
                    End If
                End If
                
                iStart = iEnd_Prior + 1
                ' If the next character is an underscore, this optional parameter was skipped
                If Mid(strKeyword, iEnd_Prior, 2) = "__" Then
                    iEnd = iStart
                    iEnd_Prior = iEnd
                    intColumns = 0
                Else
                    ' Set the start of the Columns which may come after the SourceRangeName
                    iStart = InStr(iEnd_Prior, strKeyword, "_C")
                    If iStart > 0 Then
                        ' Set the start of the numeric portion
                        iStart = iStart + 2
                        ' Get the end of the numeric portion
                        ' First look for the underscore
                        iEnd = InStr(iStart, strKeyword, "_")
                        If iEnd = 0 Then
                            ' Otherwise look for the hash
                            iEnd = InStr(iStart, strKeyword, "#")
                        End If
                        ' Update iEnd_Prior if Columns is present
                        If iEnd > 0 Then iEnd_Prior = iEnd
                        If iEnd > iStart Then
                            If IsNumeric(Mid(strKeyword, iStart, iEnd - iStart)) Then
                              intColumns = Mid(strKeyword, iStart, iEnd - iStart)
                            Else
                              intColumns = 0
                            End If
                        Else
                            'Default to 0
                            intColumns = 0
                        End If
                    Else
                        'Default to 0
                        intColumns = 0
                    End If
                End If
                
                ' Set the start of the DeleteSourceReport which may come after the SourceRangeName
                iStart = iEnd_Prior + 1
                ' If the next character is an underscore, this optional parameter was skipped
                If Mid(strKeyword, iEnd_Prior, 2) = "__" Then
                    iEnd = iStart
                    iEnd_Prior = iEnd
                    strDeleteSourceReport = "N"
                Else
                ' Get the end of the DeleteSourceReport
                    ' First look for the underscore
                    iEnd = InStr(iStart, strKeyword, "_")
            
                    If iEnd = 0 Then
                        ' Otherwise look for the hash
                        iEnd = InStr(iStart, strKeyword, "#")
                    End If
                    ' Update iEnd_Prior if DeleteSourceReport is present
                    If iEnd > 0 Then iEnd_Prior = iEnd
                    ' Parse it to find the DeleteSourceReport
                    If iEnd > iStart Then
                        strDeleteSourceReport = Mid(strKeyword, iStart, iEnd - iStart)
                    Else
                        strDeleteSourceReport = "N"
                    End If
                End If
                
                ' Set the start of the ValuesOrFormulas which may come after the SourceRangeName
                iStart = iEnd_Prior + 1
                ' If the next character is an underscore, this optional parameter was skipped
                If Mid(strKeyword, iEnd_Prior, 2) = "__" Then
                    iEnd = iStart
                    iEnd_Prior = iEnd
                    strValuesOrFormulas = "V"
                Else
                    ' Get the end of the ValuesOrFormulas
                    ' First look for the underscore
                    iEnd = InStr(iStart, strKeyword, "_")
                    If iEnd = 0 Then
                        ' Otherwise look for the hash
                        iEnd = InStr(iStart, strKeyword, "#")
                    End If
                    ' Update iEnd_Prior if ValuesOrFormulas is present
                    If iEnd > 0 Then iEnd_Prior = iEnd
                    ' Parse it to find the ValuesOrFormulas
                    If iEnd > iStart Then
                        strValuesOrFormulas = Mid(strKeyword, iStart, iEnd - iStart)
                    Else
                        strValuesOrFormulas = "V"
                    End If
                End If
                
                ' Set the start of the MyFormatShape which may come after the SourceRangeName
                iStart = iEnd_Prior + 1
                ' If the next character is an underscore, this optional parameter was skipped
                If Mid(strKeyword, iEnd_Prior, 2) = "__" Then
                    iEnd = iStart
                    iEnd_Prior = iEnd
                    strMyFormatShape = "MyFormatShape"
                Else
                ' Get the end of the MyIcon
                    ' First look for the underscore
                    iEnd = InStr(iStart, strKeyword, "_")
                    If iEnd = 0 Then
                        ' Otherwise look for the hash
                        iEnd = InStr(iStart, strKeyword, "#")
                    End If
                    ' Update iEnd_Prior if MyIcon is present
                    If iEnd > 0 Then iEnd_Prior = iEnd
                    ' Parse it to find the MyIcon
                    If iEnd > iStart Then
                        strMyFormatShape = Mid(strKeyword, iStart, iEnd - iStart)
                    Else
                        strMyFormatShape = "MyFormatShape"
                    End If
                End If
                                
                Debug.Print strKeyword
                Debug.Print "strDelimiter " & strDelimiter
                Debug.Print "strFileName " & strFileName
                Debug.Print "strSourceReport " & strSourceReport
                Debug.Print "strSourceRangeName " & strSourceRangeName
                Debug.Print "intRows " & intRows
                Debug.Print "intColumns " & intColumns
                Debug.Print "strDeleteSourceReport " & strDeleteSourceReport
                Debug.Print "strValuesOrFormulas " & strValuesOrFormulas
                Debug.Print "strMyFormatShape " & strMyFormatShape
                
                strLastGoodStatement = "TextFile parsing complete Worksheets(" + strReport + ") Range(" + rng.Address + ")"
                
                ' Sheet is protected
                If rng.Parent.ProtectContents Then
                    Call LogError("TextFile", "Sheet protected. Could not process (" & strKeyword & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                   ' Stop the loop
                    boolFound = False
                ' strSourceReport worksheet deosn't exist
                ElseIf Not WorksheetExists(strXLS, strSourceReport, True) Then
                    Call LogError("TextFile", "Invalid SourceReport. Could not process (" & strKeyword & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                   ' Stop the loop
                    boolFound = False
                ' strSourceRangeName doesn't exist
                ElseIf Not NamedRangeExists(Workbooks(strXLS).Worksheets(strSourceReport), strSourceRangeName, True, False) Then
                    Call LogError("TextFile", "Invalid SourceRangeName. Could not process (" & strKeyword & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                   ' Stop the loop
                    boolFound = False
                Else
                    ' Remove the keyword from the cell value
                    rng.Value = Replace(rng.Value, strKeyword, "")
                    ' Set the start of the source range
                    Set rngSource = Workbooks(strXLS).Worksheets(strSourceReport).Range(strSourceRangeName)
                    ' Confirm the rows
                    If intRows = 0 Then
                        ' Extend from the current row to the end of the used range
                        intRows = rngSource.Parent.UsedRange.Rows.Count - rngSource.Row + 1
                    End If
                    ' Confirm the columns
                    If intColumns = 0 Then
                        ' Extend from the current column to the end of the used range
                        intColumns = rngSource.Parent.UsedRange.Columns.Count - rngSource.Column + 1
                    End If
                    ' Set the entire source range
                    Set rngSource = rngSource.Parent.Range(rngSource.Parent.Cells(rngSource.Row, rngSource.Column), _
                        rngSource.Parent.Cells(rngSource.Row + intRows - 1, rngSource.Column + intColumns - 1))
               
                    ' Set variables required for building the text file
                    If Left(strDelimiter, 1) = "T" Then
                        strDelimiterCharacter = vbTab
                    Else
                        strDelimiterCharacter = ","
                    End If
                    iFile = FreeFile
                    strOutputFile = Workbooks(strXLS).Path & "\" & strFileName
                    
                    ' If the file exists, append the date/time stamp
                    If Dir(strOutputFile) <> "" Then
                        Call LogError("TextFile", "File " & strOutputFile & " already exists. Date/time stamp has been added to the file name.")
                        If InStrRev(strOutputFile, ".") > 2 Then
                            ' Append the date time stamp before the file extension, if there is one
                            strOutputFile = Left(strFileName, InStrRev(strOutputFile, ".") - 1) & "_" & Format(Now, "yyyymmdd_hhMMss") & _
                                    Right(strOutputFile, Len(strOutputFile) - InStrRev(strOutputFile, ".") + 1)
                        Else
                            ' If no file extension, simply append to the end
                            strOutputFile = strOutputFile & "_" & Format(Now, "yyyymmdd_hhMMss")
                        End If
                    End If
                    
                    'Open the output file
                    Open strOutputFile For Output As iFile
                    
                    ' Loop through the rows
                    For Each Row In rngSource.Rows
                        ' Loop through the cells in the rows
                        For Each Cell In Row.Cells
                            Select Case Left(strValuesOrFormulas, 1)
                                Case "F"
                                    If strDelimiterCharacter = vbTab Then
                                        strOutput = strOutput & Cell.Formula & strDelimiterCharacter
                                    Else
                                        strOutput = strOutput & Chr(34) & Cell.Formula & Chr(34) & strDelimiterCharacter
                                    End If
                                Case Else
                                    If strDelimiterCharacter = vbTab Then
                                        strOutput = strOutput & Cell.Value & strDelimiterCharacter
                                    Else
                                        strOutput = strOutput & Chr(34) & Cell.Value & Chr(34) & strDelimiterCharacter
                                    End If
                            End Select
                            
                        Next Cell
                        ' Remove the last delimiter
                        strOutput = Left(strOutput, Len(strOutput) - Len(strDelimiterCharacter))
                        
                        ' Write to the output file and reinitialize the variables
                        Print #iFile, strOutput
                        strOutput = ""
                    Next Row
                    
                    'Close the output file
                    Close iFile

                    With Workbooks(strXLS).Worksheets(strReport)
                        'Activate the sheet
                        .Activate
                        'Activate the destination range
                        .Range(rng.Address).Activate
                        ' Embed the output file
                        Set myEmbededShape = .Shapes.AddOLEObject(FileName:=strOutputFile, DisplayAsIcon:=False, Link:=False)
                    End With
                    
                    strLastGoodStatement = "TextFile created Worksheets(" + strReport + ") Range(" + rng.Address + ")"
                    
                    'Delete the output file, but only if it is a .txt file. Other file types will be locked
                    If Right(strOutputFile, 4) = ".txt" Then
                        Call KillFile(strOutputFile, True)
                    End If
                    
                    If strMyFormatShape <> "" Then
                        ' Test to see if a formatting worksheet exists
                        If WorksheetExists(strXLS, strFormatSheet, False) Then
                            ' Test to see that the format shape exists
                            If ShapeExists(Workbooks(strXLS).Worksheets(strFormatSheet), strMyFormatShape, False) Then
                                ' Set the format shape
                                Set myFormatShape = Workbooks(strXLS).Worksheets(strFormatSheet).Shapes(strMyFormatShape)
                            ElseIf strMyFormatShape <> "MyFormatShape" Then
                                'Log the error only if the non-default shape name was specified
                                Call ShapeExists(Workbooks(strXLS).Worksheets(strFormatSheet), strMyFormatShape, True)
                            End If
                        ElseIf strMyFormatShape <> "MyFormatShape" Then
                            'Log the error only if the non-default shape name was specified
                            Call WorksheetExists(strXLS, strFormatSheet, True)
                         End If
                    End If
                    
                    With myEmbededShape
                        ' If MyFormatShape exists, use it to format myEmbededShape
                        If Not myFormatShape Is Nothing Then
                            .Fill.ForeColor = myFormatShape.Fill.ForeColor
                            .LockAspectRatio = 0 ' Turn off so we can change the height and width
                            .Height = myFormatShape.Height
                            .Width = myFormatShape.Width
                            .Line.ForeColor = myFormatShape.Line.ForeColor
                            .Line.BackColor = myFormatShape.Line.BackColor
                            .Line.Weight = myFormatShape.Line.Weight
                            .Line.Style = myFormatShape.Line.Style
                        Else
                        ' Otherwise, use default values
                            .Fill.ForeColor.RGB = 65535 ' Yellow
                            .LockAspectRatio = 0 ' Turn off so we can change the height and width
                            .Height = 150
                            .Width = 300
                            .Line.ForeColor.RGB = 0
                            .Line.BackColor.RGB = 0
                            .Line.Weight = 1
                            .Line.Style = 1
                        End If
                    End With
                
                    strLastGoodStatement = "TextFile formating completed Worksheets(" + strReport + ") Range(" + rng.Address + ")"
                
                    If Left(strDeleteSourceReport, 1) = "Y" Then
                        Application.Workbooks(strXLS).Worksheets(strSourceReport).Delete
                        strLastGoodStatement = "Application.Workbooks(" + strXLS + ").Worksheets(" + strSourceReport + ").Delete"
                    End If
                
                End If
            Else
                ' Stop the loop
                boolFound = False
            End If
        End With
        
    Loop
    
ExitProcedure:
    Set rng = Nothing
    Set rngSource = Nothing
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("TextFile", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "TextFile"
            #End If
            
    End Select
    Resume ExitProcedure
    

End Sub

Private Sub RowsAutoFit(ByVal rngTarget As Range)
' 2010-04-30 JPD This overcomes limitation of merged cells not allowing Rows.AutoFit

' From http://help.lockergnome.com/office/Merged-cells-Autofit-row-height--ftopict917862.html
' Here is code from Greg Wilson.

    Dim NewRwHt As Single
    Dim cWdth As Single, MrgeWdth As Single
    Dim c As Range, cc As Range
    Dim ma As Range
    
    With rngTarget
        If .MergeCells And .WrapText Then
            Set c = .Cells(1, 1)
            cWdth = c.ColumnWidth
            Set ma = c.MergeArea
            For Each cc In ma.Cells
                MrgeWdth = MrgeWdth + cc.ColumnWidth
            Next
            ma.MergeCells = False
            c.ColumnWidth = MrgeWdth
            c.EntireRow.AutoFit
            NewRwHt = c.RowHeight
            c.ColumnWidth = cWdth
            ma.MergeCells = True
            ma.RowHeight = NewRwHt
            cWdth = 0: MrgeWdth = 0
        End If
    End With
End Sub

Public Sub BuildReport_Button_Debug()
'   2009-08-13 JPD  Created to allow building of report in Debug mode from button on Admin page
    
    Stop
    Call BuildReport_Button
End Sub

Public Function LEF_Version() As String
'   2010-02-28 JPD  Created to allow the conLEFVersion to be displayed on the Message tab of the workbook

    LEF_Version = conLEFVersion
    
End Function

Public Sub DeleteOldCSV(strPath As String, iMinutes As Double)
' 2010-12-02 JPD Created

    Dim i As Long
    Dim strFoundFile
    Dim dtFoundFile As Date
    Dim strErrorMessage As String
    
    On Error GoTo ErrorHandler
    
    ' First test the path. If it's not YSI.NetClient\, we're done
    If Right(strPath, 14) <> "YSI.NetClient\" Then
        Call LogError("DeleteOldCSV", "Invalid path: " & strPath & vbCrLf & "Old CSV files will only be deleted from the YSI.NetClient folder.")
    
    ' Then look for the presence of a file named LEF_NoDeleteOldCSV.txt which will prevent the deletion of files from this folder.
    ElseIf Dir(strPath & "LEF_NoDeleteOldCSV.txt") <> "" Then
        ' We are done, but there is no error to log
    Else
        ' Otherwise, loop through the CSV files and delete any that are more than iMinutes old
        strFoundFile = Dir(strPath & "*.csv")
        Do While Len(strFoundFile) > 0
            dtFoundFile = FileDateTime(strPath & strFoundFile)
            ' Debug.Print strFoundFile & " - " & dtFoundFile & "    " & DateAdd("n", iMinutes, dtFoundFile)
            If DateAdd("n", iMinutes, dtFoundFile) < Now() Then
                ' Debug.Print strFoundFile & " - " & dtFoundFile & " File will be deleted"
                Call KillFile(strPath & strFoundFile, True)
            End If
            strFoundFile = Dir
        Loop
    End If

ExitProcedure:
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("DeleteOldCSV", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "TextFile"
            #End If
            
    End Select
    Resume ExitProcedure
    

End Sub

Public Sub DeleteMyCSV()
'   2014-09-15 JPD  Created
'   2014-11-03 JPD  Changed KillFile BoolLogError to False

    Dim i As Long
    Dim strFoundFile
    Dim dtFoundFile As Date
    Dim strErrorMessage As String
    Dim strPath As String
    
    On Error GoTo ErrorHandler
    
    strPath = myCSVPath & "\"
    
    ' Look for the presence of a file named LEF_NoDeleteOldCSV.txt which will prevent the deletion of files from this folder.
    If Dir(strPath & "LEF_NoDeleteOldCSV.txt") <> "" Then
        ' We are done, but there is no error to log
    Else
        ' Otherwise, find and delete the CSV file(s) related to this report
        ' First, myCSV
        strFoundFile = Dir(strPath & myCSV)
        If Len(strFoundFile) > 0 Then
            Call KillFile(strPath & strFoundFile, False)
        End If
        
        ' Second, the non-UTF8 version of myCSV, if such a file exists
        strFoundFile = Dir(strPath & Replace(myCSV, "UTF-8.csv", ".csv"))
        If Len(strFoundFile) > 0 Then
            Call KillFile(strPath & strFoundFile, False)
        End If
    
    End If

ExitProcedure:
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("DeleteMyCSV", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "DeleteMyCSV"
            #End If
            
    End Select
    Resume ExitProcedure
    

End Sub

Public Sub GetTitles(strXLS As String, strReport As String)
    ' 2011-05-13 JPD Created to make it easier to display titles, including filter selections, in the report tabs
    
    ' This finds and replaces keywords #Titles_MyTitleType_MyFilterCaption#
    ' 1. Must start with #Titles (this syntax is reserved for the Titles function)
    ' 2. MyTitleType (Required)
    '    - All/AllExcept1/Filters/Filters1Row/FilterTitle/FilterValue/XX
    '       - All = All title rows from CSV file, separated by CHR(10)
    '         - 0 can be used in place of All
    '       - AllExcept1 = All title rows from CSV file except the first title row, separated by CHR(10)
    '         - This allows the report title to be in a separate cell, with different formatting
    '         - -1 can be used in place of AllExcept1
    '       - Filters = All title rows related to filters from CSV file, separated by CHR(10)
    '       - Filters1Row = All title rows related to filters from CSV file, separated by " AND "
    '       - FilterTitle = A single title row for the filter specified by MyFilterCaption
    '       - FilterValue = The selected filter value(s) for the filter specified by MyFilterCaption
    '       - XX = A single title row for the filter specified by MyFilterCaption
    '         - XX must be a number between 1 and 1000
    ' 3. MyFilterCaption
    '    -  Caption from the filter section of the report (this is what the user sees on the filter)
    '    -  Cannot contain an underscore, so use a ^ as a placeholder for underscores in the caption name
    '    -  Cannot contain a #, so use a pipe (|) as a placeholder for hashes
    '    -  Required when MyTitleType is FilterTitle or FilterValue
    ' 4. Must end with # and have not other # characters in this keyword
    ' 5. Titles keywords can be located anywhere in the cell's text
    ' 6. Multiple Titles keywords in the same cell will be processed from left to right
    ' 7. Titles starting with a chevron (<) are assumed to be LEF instructions and are not returned
      
    On Error GoTo ErrorHandler
    
    Dim rng As Range
    Dim boolFound As Boolean
    Dim iStart As Long
    Dim iEnd As Long
    Dim iEnd_Prior As Long
    Dim strPrefix As String
    Dim strKeyword As String
    Dim strTitleType As String
    Dim strFilterCaption As String
    Dim strErrorMessage As String
    Dim strFoundTitle As String
    Dim intTitleRow As Long
    
    ' Initialize variable
    boolFound = True
    strPrefix = "#Titles_" ' Hard coded for this procedure
    
    Do While boolFound = True
        With Workbooks(strXLS).Worksheets(strReport).UsedRange
            Set rng = .Find(What:="*" + strPrefix + "*#*", _
                            after:=.Cells(.Cells.Count), _
                            LookIn:=xlFormulas, _
                            LookAt:=xlWhole, _
                            SearchOrder:=xlByRows, _
                            SearchDirection:=xlNext, _
                            MatchCase:=False)
            If Not rng Is Nothing Then
                ' Get the start of the keyword
                iStart = InStr(1, rng.Value, strPrefix)
                ' Get the end of the keyword
                iEnd = InStr(iStart + Len(strPrefix), rng.Value, "#")
                ' Parse it to find the keyword.
                strKeyword = Mid(rng.Value, iStart, iEnd - iStart + 1)
                ' Set the start of the FilterCaption which come after the Prefix
                iStart = Len(strPrefix) + 1
                ' Get the end of the TitleType
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if TitleType is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the TitleType
                If iEnd > iStart Then
                    strTitleType = Mid(strKeyword, iStart, iEnd - iStart)
                Else
                    strTitleType = ""
                End If
                ' Set the start of the FilterPart which come after the TitleType
                iStart = iEnd + 1
                ' Get the end of the FilterPart
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if FilterCaption is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the FilterCaption
                If iEnd > iStart Then
                    strFilterCaption = Mid(strKeyword, iStart, iEnd - iStart)
                Else
                    strFilterCaption = ""
                End If
                
                ' If the sheet is locked then exit. There is nothing we can do.
                If Workbooks(strXLS).Worksheets(strReport).ProtectContents Then
                    ' Stop the loop
                    boolFound = False
                    Call LogError("GetTitles", "Sheet protected. Could not process (" & strKeyword & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                End If

                ' Get values for the requested title
                With Application.Workbooks(strXLS).Worksheets("LEFTitles")
                    If strTitleType = "All" Then
                        strFoundTitle = .Cells(1, 2).Value  ' All is in cell B1
                    ElseIf strTitleType = "AllExcept1" Then
                        strFoundTitle = .Cells(1, 3).Value  ' AllExcept1 is in cell C1
                    ElseIf strTitleType = "Filters" Then
                        strFoundTitle = .Cells(1, 4).Value  ' Filters is in cell D1
                    ElseIf strTitleType = "Filters1Row" Then
                        strFoundTitle = .Cells(1, 5).Value  ' Filter1Row is in cell E1
                    ElseIf strTitleType = "FilterTitle" Or strTitleType = "FilterValue" Then
                        ' Find the row with this strFilterCaption. Substitute _ for ^ in strFilterCaption.
                        intTitleRow = FirstMatch(Replace(Replace(strFilterCaption, "^", "_"), "|", "#") & "=*", strXLS, "LEFTitles", "A:A")
                        If intTitleRow = -1 Then
                            strFoundTitle = ""
                        ElseIf strTitleType = "FilterTitle" Then
                            strFoundTitle = .Cells(intTitleRow, 1).Value  ' Titles are in column A
                        ElseIf strTitleType = "FilterValue" Then
                            strFoundTitle = .Cells(intTitleRow, 3).Value  ' Values are in column C
                        End If
                    ElseIf IsNumeric(strTitleType) Then
                        intTitleRow = strTitleType
                        If intTitleRow = 0 Then
                            strFoundTitle = .Cells(1, 2).Value  ' All is in cell B1
                        ElseIf intTitleRow = -1 Then
                            strFoundTitle = .Cells(1, 3).Value  ' AllExcept1 is in cell C1
                        ElseIf intTitleRow >= 1 And intTitleRow <= 100 Then
                            strFoundTitle = .Cells(intTitleRow, 1).Value  ' Titles are in column A
                        Else
                            strFoundTitle = ""
                        End If
                    End If
                End With
                
                ' Replace the keyword with the found title
                rng.Value = Replace(rng.Value, strKeyword, strFoundTitle)
                strFoundTitle = ""

            Else
                ' Stop the loop
                boolFound = False
            End If
        End With
    Loop
    
ExitProcedure:
    Set rng = Nothing
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("GetTitles", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "GetTitles"
            #End If
            
    End Select
    Resume ExitProcedure

End Sub

Public Sub HyperlinkCreate(strXLS As String, strReport As String, strPrefix As String)
    ' 2011-05-13 JPD Created to allow drill down
    
    ' strPrefix = "#Hyperlink_"
    
    ' This finds and replaces keywords #Hyperlink_MyAddress_MySubAddress_MyScreenTip_MyTextToDisplay#
    ' 1. Must start with #Hyperlink (this syntax is reserved for the HyperlinkCreate function)
    ' 2. MyAddress (Required)
    '    - The address of the hyperlink
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores
    '    - Cannot contain a #, so use a pipe (|) as a placeholder for hashes
    '      - Web site example: http://www.smartbizniz.com/
    '      - Link to a cell in the same workbook: leave blank
    '      - Email example: mailto:John@SmartBizNiz.com?subject=MySubject
    '      - File example: c:\data\MyFile.pdf
    ' 3. MySubAddress (Optional)
    '    - The cell reference, named range or bookmark you are navigating to.
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores
    '    - Cannot contain a #, so use a pipe (|) as a placeholder for hashes
    '      - Cell in a worksheet example: MyWorksheet!A15
    '      - Named Range with worksheet scope example: MyWorksheet!MyLocalNamedRange
    '      - Named Range with workbook scope example: MyGlobalNamedRange
    ' 4. MyScreenTip (Optional)
    '    - The screen tip to be displayed when the mouse pointer is paused over the hyperlink.
    '    - Cannot contain a #, so use a pipe (|) as a placeholder for hashes
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores
    '      - Example: Click here to go to www.SmartBizNiz.com
    ' 5. MyTextToDisplay (Optional)
    '    - The text to be displayed for the hyperlink. This is what is visible.
    '    - Cannot contain a #, so use a pipe (|) as a placeholder for hashes
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores
    '      - Any formatting should be handled here with #FC
    ' 6. Must end with # and have not other # characters in this keyword
    '    - Pipe (|) can be used as a placeholder for hashes for embedded keywords
    ' 5. Hyperlink keywords can be located anywhere in the cell's text
    ' 6. Multiple Hyperlink keywords in the same cell will be processed from left to right
      
    On Error GoTo ErrorHandler
    
    Dim rng As Range
    Dim boolFound As Boolean
    Dim iStart As Long
    Dim iEnd As Long
    Dim iEnd_Prior As Long
    Dim strKeyword As String
    Dim strAddress As String
    Dim strSubAddress As String
    Dim strScreenTip As String
    Dim strTextToDisplay As String
    Dim strErrorMessage As String
    Dim strFoundTitle As String
    Dim intTitleRow As Long
    
    ' Initialize variable
    boolFound = True
    
    Do While boolFound = True
        With Workbooks(strXLS).Worksheets(strReport).UsedRange
            Set rng = .Find(What:="*" + strPrefix + "*#*", _
                            after:=.Cells(.Cells.Count), _
                            LookIn:=xlFormulas, _
                            LookAt:=xlWhole, _
                            SearchOrder:=xlByRows, _
                            SearchDirection:=xlNext, _
                            MatchCase:=False)
            If Not rng Is Nothing Then
                ' Get the start of the keyword
                iStart = InStr(1, rng.Value, strPrefix)
                ' Get the end of the keyword
                iEnd = InStr(iStart + Len(strPrefix), rng.Value, "#")
                ' Parse it to find the keyword.
                strKeyword = Mid(rng.Value, iStart, iEnd - iStart + 1)
                ' Set the start of the Address which come after the Prefix
                iStart = Len(strPrefix) + 1
                ' Get the end of the Address
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if Address is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the Address
                If iEnd > iStart Then
                    strAddress = Replace(Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_"), "|", "#")
                Else
                    strAddress = ""
                End If
                ' Set the start of the SubAddress which come after the Address
                iStart = iEnd + 1
                ' Get the end of the SubAddress
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if SubAddress is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the SubAddress
                If iEnd > iStart Then
                    strSubAddress = Replace(Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_"), "|", "#")
                Else
                    strSubAddress = ""
                End If
                ' Set the start of the ScreenTip which come after the SubAddress
                iStart = iEnd + 1
                ' Get the end of the ScreenTip
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if ScreenTip is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the ScreenTip
                If iEnd > iStart Then
                    strScreenTip = Replace(Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_"), "|", "#")
                Else
                    strScreenTip = ""
                End If
                ' Set the start of the TextToDisplay which come after the ScreenTip
                iStart = iEnd + 1
                ' Get the end of the TextToDisplay
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if TextToDisplay is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the TextToDisplay
                If iEnd > iStart Then
                    strTextToDisplay = Replace(Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_"), "|", "#")
                Else
                    strTextToDisplay = ""
                End If
                
                ' If the sheet is locked then exit. There is nothing we can do.
                If Workbooks(strXLS).Worksheets(strReport).ProtectContents Then
                    ' Stop the loop
                    boolFound = False
                    Call LogError("HyperlinkCreate", "Sheet protected. Could not process (" & strKeyword & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                End If

                ' Create the hyperlink
                With Workbooks(strXLS).Worksheets(strReport)
                    .Hyperlinks.Add Anchor:=.Range(rng.Address), _
                        Address:=strAddress, _
                        SubAddress:=strSubAddress, _
                        ScreenTip:=strScreenTip, _
                        TextToDisplay:=strTextToDisplay
                End With
           
                ' Replace the keyword, but it should already be gone
                rng.Value = Replace(rng.Value, strKeyword, "")

            Else
                ' Stop the loop
                boolFound = False
            End If
        End With
    Loop
    
ExitProcedure:
    Set rng = Nothing
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("HyperlinkCreate", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "HyperlinkCreate"
            #End If
            
    End Select
    Resume ExitProcedure

End Sub

Public Sub PictureInsert(strXLS As String, strReport As String, strPrefix As String)
    ' 2011-05-13 JPD Created to allow inserting logos or other images
    ' 2011-11-09 JPD Added Alignment attribute
    ' 2013-10-23 JPD Added Placement to the Alignment attribute
    '                Adjusted code to work properly with Excel 2003
    
    ' strPrefix = "#Picture_"
    
    ' This finds and replaces keywords #Picture_MyPath_Fill,FillV,FillH,Top,Bottom,CenterV,Left,Right,CenterH#
    ' 1. Must start with #Picture (this syntax is reserved for the PictureInsert function)
    ' 2. MyPath (Required)
    '    - The path (UNC or URL) to the picture
    '    - Cannot contain an underscore, so use a ^ as a placeholder for underscores
    '    - Cannot contain a #, so use a pipe (|) as a placeholder for hashes
    '      - Local path example: C:\Data\MyLogo.jpg
    '      - Internet example: http://www.google.com/images/logos/ps_logo2.png
    ' 3. Alignment/Placement (Optional)
    '    - Alignment controls where the picture is placed relative to this cell
    '    - Defaults are Top and Left which places the picture in the upper left corner of cell
    '      - Fill = Fills this cell with the picture. Aspect Ratio is not maintained
    '      - FillV = Vertically fills this cell with the picture. Aspect Ratio is maintained
    '      - FillH = Horizonatally fills this cell with the picture. Aspect Ratio is maintained
    '      - Bottom = Bottom of picture aligns with the bottom of this cell
    '      - CenterV = Picture aligns vertically with the center of this cell
    '      - Right = Picture is aligned to right edge of this cell
    '      - CenterH = Picture aligns horizontally with the center of this cell
    '    - Placement controls where the picture is positioned when cells are moved
    '    - Default is Move = Object is moved with the cells
    '      - FreeFloating = Object is free floating
    '      - MoveAndSize = Object is moved and sized with the cells.
    '      - Move = Object is moved with the cells. This is the default.
    ' 4. Must end with # and have not other # characters in this keyword
    '    - Pipe (|) can be used as a placeholder for hashes for embedded keywords
    ' 5. Picture keyword can be located anywhere in the cell's text
    ' 6. Multiple Picture keywords in the same cell will be processed from left to right
      
    On Error GoTo ErrorHandler
    
    Dim rng As Range
    Dim boolFound As Boolean
    Dim iStart As Long
    Dim iEnd As Long
    Dim iEnd_Prior As Long
    Dim strKeyword As String
    Dim strPath As String
    Dim strAlignment As String
    Dim objPicture As Object
    Dim dTop As Double
    Dim dLeft As Double
    Dim dHeight As Double
    Dim dWidth As Double
    Dim dPictureHeight As Double
    Dim dPictureWidth As Double
    Dim strErrorMessage As String
    Dim strFoundTitle As String
    
    ' Initialize variable
    boolFound = True
    
    Do While boolFound = True
        With Workbooks(strXLS).Worksheets(strReport).UsedRange
            Set rng = .Find(What:="*" + strPrefix + "*#*", _
                            after:=.Cells(.Cells.Count), _
                            LookIn:=xlFormulas, _
                            LookAt:=xlWhole, _
                            SearchOrder:=xlByRows, _
                            SearchDirection:=xlNext, _
                            MatchCase:=False)
            If Not rng Is Nothing Then
                ' Get the start of the keyword
                iStart = InStr(1, rng.Value, strPrefix)
                ' Get the end of the keyword
                iEnd = InStr(iStart + Len(strPrefix), rng.Value, "#")
                ' Parse it to find the keyword.
                strKeyword = Mid(rng.Value, iStart, iEnd - iStart + 1)
                ' Set the start of the Path which comes after the Prefix
                iStart = Len(strPrefix) + 1
                ' Get the end of the Path
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if Path is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the Path
                If iEnd > iStart Then
                    strPath = Replace(Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_"), "|", "#")
                Else
                    strPath = ""
                End If
                
                 ' Set the start of the Alignment which comes after the Path
                iStart = iEnd + 1
                ' Get the end of the Alignment
                ' First look for the underscore
                iEnd = InStr(iStart, strKeyword, "_")
                If iEnd = 0 Then
                    ' Otherwise look for the hash
                    iEnd = InStr(iStart, strKeyword, "#")
                End If
                ' Update iEnd_Prior if Alignment is present
                If iEnd > 0 Then iEnd_Prior = iEnd
                ' Parse it to find the Alignment
                If iEnd > iStart Then
                    strAlignment = Replace(Replace(Mid(strKeyword, iStart, iEnd - iStart), "^", "_"), "|", "#")
                Else
                    strAlignment = ""
                End If
               
                ' If the sheet is locked then exit. There is nothing we can do.
                If Workbooks(strXLS).Worksheets(strReport).ProtectContents Then
                    ' Stop the loop
                    boolFound = False
                    Call LogError("PictureInsert", "Sheet protected. Could not process (" & strKeyword & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                ' If the sheet is locked then exit. There is nothing we can do.
                ElseIf Len(strPath) = 0 Then
                    Call LogError("PictureInsert", "Invalid filename. Could not process (" & strKeyword & ") in Worksheets(" + strReport + ") Cells(" + rng.Address + ")")
                End If

                ' Insert the picture
                Set objPicture = Workbooks(strXLS).Worksheets(strReport).Pictures.Insert(strPath)
                
                With rng
                    ' If Fill, then set the picture width and height as needed
                    If strAlignment Like "*FillH*" Then
                        objPicture.ShapeRange.LockAspectRatio = msoTrue
                        objPicture.ShapeRange.Width = .Offset(0, .Columns.Count).Left - .Left
                    ElseIf strAlignment Like "*FillV*" Then
                        objPicture.ShapeRange.LockAspectRatio = msoTrue
                        objPicture.ShapeRange.Height = .Offset(.Rows.Count, 0).Top - .Top
                    ElseIf strAlignment Like "*Fill*" Then
                        objPicture.ShapeRange.LockAspectRatio = msoFalse
                        objPicture.ShapeRange.Width = .Offset(0, .Columns.Count).Left - .Left
                        objPicture.ShapeRange.Height = .Offset(.Rows.Count, 0).Top - .Top
                    End If
                    
                    ' Set the default values
                    dTop = .Top
                    dLeft = .Left
                    dPictureWidth = objPicture.ShapeRange.Width
                    dPictureHeight = objPicture.ShapeRange.Height
                    
                    ' Check the horizontal options
                    If strAlignment Like "*CenterH*" Then
                        dWidth = .Offset(0, .Columns.Count).Left - .Left
                        dLeft = dLeft + dWidth / 2 - dPictureWidth / 2
                        If dLeft < 1 Then dLeft = 1
                    ElseIf strAlignment Like "*Right*" Then
                        dWidth = .Offset(0, .Columns.Count).Left - .Left
                        dLeft = dLeft + dWidth - dPictureWidth
                        If dLeft < 1 Then dLeft = 1
                    End If
                    
                    ' Check the vertical options
                    If strAlignment Like "*CenterV*" Then
                        dHeight = .Offset(1, 0).Top - .Top
                        dTop = dTop + dHeight / 2 - dPictureHeight / 2
                        If dTop < dTop Then dTop = 1
                    ElseIf strAlignment Like "*Bottom*" Then
                        dHeight = .Offset(1, 0).Top - .Top
                        dTop = dTop + dHeight - dPictureHeight
                        If dTop < dTop Then dTop = 1
                    End If
                End With
                    
                ' Position the picture
                With objPicture.ShapeRange
                    .Top = dTop
                    .Left = dLeft
                End With
                
                ' Check the Placement options
                If strAlignment Like "*FreeFloating*" Then
                    objPicture.Placement = xlFreeFloating       ' Object is free floating.
                ElseIf strAlignment Like "*MoveAndSize*" Then
                    objPicture.Placement = xlMoveAndSize        ' Object is moved and sized with the cells.
                ElseIf strAlignment Like "*Move*" Then
                    objPicture.Placement = xlMove               ' Object is moved with the cells. This is the default.
                End If
                
                ' Replace the keyword
                rng.Value = Replace(rng.Value, strKeyword, "")

            Else
                ' Stop the loop
                boolFound = False
            End If
        End With
    Loop
    
ExitProcedure:
    Set rng = Nothing
    Set objPicture = Nothing
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("PictureInsert", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "PictureInsert"
            #End If
            
    End Select
    Resume ExitProcedure

End Sub

Public Sub PageHeaderFooter(strXLS As String, strReport As String, strPrefix As String)
    ' 2013-10-06 JPD Created to allow setting custom headers and footers
    ' 2013-11-12 JPD Tweaked
    ' 2014-11-20 JPD Fixed bug that caused errors for anything except for strLeftFooter
    
    ' strPrefix = "", "After", "Protect", "Final"
    
    ' This finds named ranges PageXXXXstrPrefix_Position and sets the Header/Footer = value of fist column in each row of the named range
    ' - Example of named ranges: PageHeader_Left, PageHeaderAfter_Left, PageFooterFinal_Center
    ' 1. XXXX = Header or Footer
    ' 2. strPrefix (Optional)
    '    - Determines when this process runs
    '    - "", "After", "Protect", "Final"
    ' 3. Position (Required)
    '    - Left, Right, Center
    '
    ' See the following resources for sample syntax for formatting header and footer text
    ' http://www.cpearson.com/excel/headfoot.htm
    ' http://support.microsoft.com/kb/142136
    ' http://msdn.microsoft.com/en-us/library/office/ff822794(v=office.14).aspx
      
    On Error GoTo ErrorHandler
    
    Dim strErrorMessage As String
    Dim ws As Worksheet
    Dim strLeftHeader As String
    Dim strCenterHeader As String
    Dim strRightHeader As String
    Dim strLeftFooter As String
    Dim strCenterFooter As String
    Dim strRightFooter As String
    Dim strValue As String
    Dim i As Integer
    
    Set ws = Workbooks(strXLS).Worksheets(strReport)
    
    strLeftHeader = "PageHeader" + strPrefix + "_Left"
    strCenterHeader = "PageHeader" + strPrefix + "_Center"
    strRightHeader = "PageHeader" + strPrefix + "_Right"
    
    strLeftFooter = "PageFooter" + strPrefix + "_Left"
    strCenterFooter = "PageFooter" + strPrefix + "_Center"
    strRightFooter = "PageFooter" + strPrefix + "_Right"
    
    If NamedRangeExists(ws, strLeftHeader, False) Then
        strValue = ""
        For i = 1 To ws.Range(strLeftHeader).Rows.Count
            strValue = strValue & ws.Range(strLeftHeader).Cells(i, 1).Value
            If i < ws.Range(strLeftHeader).Rows.Count Then
                strValue = strValue & vbCr
            End If
        Next i
        ws.PageSetup.LeftHeader = strValue
    End If
    
    If NamedRangeExists(ws, strCenterHeader, False) Then
        strValue = ""
        For i = 1 To ws.Range(strCenterHeader).Rows.Count
            strValue = strValue & ws.Range(strCenterHeader).Cells(i, 1).Value
            If i < ws.Range(strCenterHeader).Rows.Count Then
                strValue = strValue & vbCr
            End If
        Next i
        ws.PageSetup.CenterHeader = strValue
    End If
    
    If NamedRangeExists(ws, strRightHeader, False) Then
        strValue = ""
        For i = 1 To ws.Range(strRightHeader).Rows.Count
            strValue = strValue & ws.Range(strRightHeader).Cells(i, 1).Value
            If i < ws.Range(strRightHeader).Rows.Count Then
                strValue = strValue & vbCr
            End If
        Next i
        ws.PageSetup.RightHeader = strValue
    End If
    
    If NamedRangeExists(ws, strLeftFooter, False) Then
        strValue = ""
        For i = 1 To ws.Range(strLeftFooter).Rows.Count
            strValue = strValue & ws.Range(strLeftFooter).Cells(i, 1).Value
            If i < ws.Range(strLeftFooter).Rows.Count Then
                strValue = strValue & vbCr
            End If
        Next i
        ws.PageSetup.LeftFooter = strValue
    End If
    
    If NamedRangeExists(ws, strCenterFooter, False) Then
        strValue = ""
        For i = 1 To ws.Range(strCenterFooter).Rows.Count
            strValue = strValue & ws.Range(strCenterFooter).Cells(i, 1).Value
            If i < ws.Range(strCenterFooter).Rows.Count Then
                strValue = strValue & vbCr
            End If
        Next i
        ws.PageSetup.CenterFooter = strValue
    End If
    
    If NamedRangeExists(ws, strRightFooter, False) Then
        strValue = ""
        For i = 1 To ws.Range(strRightFooter).Rows.Count
            strValue = strValue & ws.Range(strRightFooter).Cells(i, 1).Value
            If i < ws.Range(strRightFooter).Rows.Count Then
                strValue = strValue & vbCr
            End If
        Next i
        ws.PageSetup.RightFooter = strValue
    End If
   
ExitProcedure:
    
    Exit Sub
    
ErrorHandler:
    Select Case Err.Number
        Case Else
            strErrorMessage = Err.Number & ": " & Err.Description & vbCrLf & _
                "Last good statement: " & strLastGoodStatement
                
            Call LogError("PageHeaderFooter", strErrorMessage)
            
            #If DEBUG_ Then
                'Message boxes only in DEBUG mode
                MsgBox strErrorMessage, vbOKOnly, "PageHeaderFooter"
            #End If
            
    End Select
    Resume ExitProcedure

End Sub

